"""
Test the parsers that interacte with the AiiDA database.
Check if various AiiDA types are created correctly.
"""

from __future__ import absolute_import
import six
import pytest
from aiida_castep.common import OUTPUT_LINKNAMES

ln_name = OUTPUT_LINKNAMES

folders = ["H2-geom", "O2-geom-spin", "Si-geom-stress", "N2-md"]


def test_parsing_base(
        new_database,
        db_test_app,
        generate_calc_job_node,
        generate_parser,
        h2_calc_inputs,
):
    """
    Test basic parsing
    """

    node = generate_calc_job_node(
        'castep.castep',
        'H2-geom',
        inputs=h2_calc_inputs,
    )
    parser = generate_parser('castep.castep')
    results, return_node = parser.parse_from_node(node, store_provenance=False)
    assert return_node.exit_status == 0

    calc_energy = results['output_parameters'].get_dict()['total_energy']
    ref = -31.69654969917
    assert calc_energy == ref


def test_parse_warnings(
        new_database,
        db_test_app,
        generate_calc_job_node,
        generate_parser,
        h2_calc_inputs,
):
    """
    Test basic parsing
    """
    from aiida_castep.common import EXIT_CODES_SPEC as CODES

    node = generate_calc_job_node(
        'castep.castep',
        'H2-geom',
        inputs=h2_calc_inputs,
    )
    parser = generate_parser('castep.castep')

    folder = node.outputs.retrieved
    content_orig = folder.get_object_content('aiida.castep').split('\n')
    content = content_orig[:-20]
    content.append('Insufficient time for another iteration')
    with node.outputs.retrieved.open('aiida.castep', 'w') as fh:
        fh.write('\n'.join(content))
    results, return_node = parser.parse_from_node(node, store_provenance=False)
    assert return_node.exit_status == CODES['ERROR_TIMELIMIT_REACHED'][0]

    content.append(
        'SCF cycles performed but system has not reached the groundstate')
    with node.outputs.retrieved.open('aiida.castep', 'w') as fh:
        fh.write('\n'.join(content))
    results, return_node = parser.parse_from_node(node, store_provenance=False)
    assert return_node.exit_status == CODES['ERROR_SCF_NOT_CONVERGED'][0]

    node.outputs.retrieved.delete_object('aiida.castep', force=True)
    results, return_node = parser.parse_from_node(node, store_provenance=False)
    assert return_node.exit_status == CODES['ERROR_NO_OUTPUT_FILE'][0]


def test_parse_errs(
        new_database,
        db_test_app,
        generate_calc_job_node,
        generate_parser,
        h2_calc_inputs,
):
    """
    Test basic parsing
    """
    from aiida_castep.common import EXIT_CODES_SPEC as CODES
    from io import StringIO

    node = generate_calc_job_node(
        'castep.castep',
        'H2-geom',
        inputs=h2_calc_inputs,
    )
    parser = generate_parser('castep.castep')

    folder = node.outputs.retrieved

    error_string = u'Error Message\nError'
    err_handle = StringIO(error_string)
    folder.put_object_from_filelike(err_handle, 'aiida.0001.err', force=True)
    results, return_node = parser.parse_from_node(node, store_provenance=False)
    assert return_node.exit_status == CODES['ERROR_CASTEP_ERROR'][0]


def test_parsing_geom(new_database, db_test_app, generate_calc_job_node,
                      generate_parser, h2_calc_inputs):
    """
    Test if geom is converted to trajectory data
    """
    from aiida_castep.calculations.castep import CastepCalculation
    node = generate_calc_job_node('castep.castep',
                                  'H2-geom',
                                  inputs=h2_calc_inputs,
                                  computer=db_test_app.localhost)
    parser = generate_parser('castep.castep')
    results, _ = parser.parse_from_node(node, store_provenance=False)
    assert 'forces' in results['output_trajectory'].get_arraynames()


@pytest.mark.parametrize('output_folder', folders)
def test_parser_retrieved(db_test_app, output_folder, generate_parser,
                          generate_calc_job_node, sto_calc_inputs):
    """
    Iterate through internal test cases.
    Check if the results are parsed correctly.
    """
    from aiida_castep.tests.utils import get_x2_structure

    common_keys = [
        "cells", "positions", "forces", "symbols", "geom_total_energy"
    ]
    md_keys = [
        "hamilt_energy", "kinetic_energy", "velocities", "temperatures",
        "times"
    ]
    geom_keys = ["geom_enthalpy"]

    inputs = sto_calc_inputs

    if 'O2' in output_folder:
        xtemp = 'O'
    elif 'Si' in output_folder:
        xtemp = 'Si'
    elif 'N2' in output_folder:
        xtemp = 'N'
    elif 'H2' in output_folder:
        xtemp = 'H'

    # Swap the correct structure to allow desort to work
    inputs.structure = get_x2_structure(xtemp)
    parser = generate_parser('castep.castep')
    node = generate_calc_job_node('castep.castep', output_folder, inputs)

    out, _ = parser.parse_from_node(node, store_provenance=False)

    out_structure = out[ln_name['structure']]
    out_param_dict = out[ln_name['results']].get_dict()
    out_traj = out[ln_name['trajectory']]

    assert "total_energy" in out_param_dict
    assert "unit_energy" in out_param_dict
    assert out_param_dict["unit_energy"] == "eV"
    # Check if the label is correctly copied
    assert node.inputs.structure.label == out_structure.label

    # Check the length of sites are consistent
    assert len(out_structure.sites) == len(out_traj.symbols)

    for k in common_keys:
        assert k in out_traj.get_arraynames()

    if output_folder == "O2-geom-spin" or output_folder == "Si-geom-stress":
        bands = out.get(ln_name['bands'], None)
        assert bands is not None

        # Check if spins are handled correctly
        assert bands.get_attribute('nspins') in [1, 2]
        if bands.get_attribute('nspins') == 1:
            assert bands.get_attribute('nkpts') == len(bands.get_bands())
        elif bands.get_attribute('nspins') == 2:
            assert bands.get_attribute('nkpts') == len(bands.get_bands()[0])

        for k in geom_keys:
            assert k in out_traj.get_arraynames()

    if output_folder == "N2-md":
        for k in md_keys:
            assert k in out_traj.get_arraynames()



"""
Storing OTFG configuration as Data nodes
"""
from __future__ import absolute_import
from __future__ import print_function
import re
from aiida.orm import Data, Group, QueryBuilder
from aiida.common.utils import classproperty
from aiida.common import ValidationError
from .utils import split_otfg_entry
from .usp import UspData
import six

OLD_OTFGGROUP_TYPE = "data.castep.otfg.family"
OTFGGROUP_TYPE = "castep.otfg"


class OTFGGroup(Group):
    """Class representing an OTFGGroup"""


def migrate_otfg_family():
    """Migrate the old OTFG families to new families"""
    old_types = [OLD_OTFGGROUP_TYPE, "data.castep.usp.family"]
    q = QueryBuilder()
    q.append(Group, filters={'type_string': {'in': old_types}})

    migrated = []
    created = []
    for (old_group, ) in q.iterall():
        new_group, created = OTFGGroup.objects.get_or_create(
            label=old_group.label, description=old_group.description)
        new_group.add_nodes(list(old_group.nodes))
        new_group.store()
        migrated.append(new_group.label)
        if created:
            print("Created new style Group for <{}>".format(old_group.label))
        else:
            print(("Adding nodes to existing group <{}>".format(
                old_group.label)))

    return


def upload_otfg_family(entries,
                       group_label,
                       group_description,
                       stop_if_existing=True):
    """
    Set a family for the OTFG pseudo potential strings
    """
    from aiida.common import UniquenessError, NotExistent
    from aiida.orm.querybuilder import QueryBuilder
    #from aiida.common import aiidalogger

    # Try to retrieve a group if it exists
    try:
        group = OTFGGroup.get(label=group_label)
        group_created = False
    except NotExistent:
        group = OTFGGroup(label=group_label, )
        group_created = True

    group.description = group_description

    otfg_and_created = []
    nentries = len(entries)

    for entry in entries:
        # Add it if it is just one existing data
        if isinstance(entry, OTFGData):
            element, setting = entry.element, entry.string
        elif isinstance(entry, str):
            element, setting = split_otfg_entry(entry)
        elif isinstance(entry, UspData):
            element, setting = entry.element, entry.md5sum

        qb = QueryBuilder()
        qb.append(OTFGData,
                  filters={
                      'attributes.otfg_entry': {
                          "==": setting
                      },
                      'attributes.element': {
                          "==": element
                      }
                  })
        existing_otfg = qb.first()

        # Try find Usp data
        if existing_otfg is None:

            qb = QueryBuilder()
            qb.append(UspData,
                      filters={
                          'attributes.md5sum': {
                              "==": setting
                          },
                          'attributes.element': {
                              "==": element
                          }
                      })
            existing_otfg = qb.first()

        # Act based on wether the data exists
        if existing_otfg is None:

            if isinstance(entry, OTFGData):
                otfg_and_created.append((entry, True))
            elif isinstance(entry, str):
                otfg, created = OTFGData.get_or_create(entry,
                                                       use_first=True,
                                                       store_otfg=False)
                otfg_and_created.append((otfg, created))
            elif isinstance(entry, UspData):
                otfg_and_created.append((entry, True))

        else:
            if stop_if_existing:
                raise ValidationError(
                    "A OTFG group cannot be added when stop_if_existing is True"
                )
            existing_otfg = existing_otfg[0]
            otfg_and_created.append((existing_otfg, False))

    # Check for unique for the complete group
    elements = [(i[0].element, i[0].string) for i in otfg_and_created]

    # Add other entries for the list to check
    if not group_created:
        for aiida_n in group.nodes:
            if not isinstance(aiida_n, (OTFGData, UspData)):
                print(("Warning: unsupported node: {}".format(aiida_n)))
                continue
            elements.append((aiida_n.element, aiida_n.string))

    # Discard duplicated pairs
    elements = set(elements)
    elements_names = [e[0] for e in elements]

    # Check the uniqueness of the complete group
    if not len(elements_names) == len(set(elements_names)):
        duplicates = set(
            [x for x in elements_names if elements_names.count(x) > 1])
        dup_string = ", ".join(duplicates)
        raise UniquenessError("More than one Nodes found for the elements: " +
                              dup_string + ".")

    # If we survive here uniqueness is fine

    # Save the group - note we have not added the nodes yet
    if group_created:
        group.store()

    # Save the OTFG in the database if necessary and add them to the group

    for otfg, created in otfg_and_created:
        if created:
            otfg.store()
        else:
            pass

    nodes_add = [otfg for otfg, created in otfg_and_created]
    nodes_new = [otfg for otfg, created in otfg_and_created if created is True]
    group.add_nodes(nodes_add)

    return nentries, len(nodes_new)


class OTFGData(Data):
    """
    Class representing an OTFG configuration

    attributes:
    string: string to be put into the cell file
    element: element that this setting is for - may not exist if we are dealing with a library.
    """
    def __init__(self, **kwargs):
        """
        Store a string for on-the-fly generation of pseudopoentials

        :param otfg_entry str: a string specifying the generation.
        The element this  potential is for can also be included.
        For example: 'O 2|1.1|15|18|20|20:21(qc=7)'
        """
        otfg_entry = kwargs.pop('otfg_entry', None)
        super(OTFGData, self).__init__(**kwargs)
        if otfg_entry:
            element, entry = split_otfg_entry(otfg_entry)
            self.set_string(entry)
            if element:
                self.set_element(element)

    @classmethod
    def get_or_create(cls, otfg_entry, use_first=False, store_otfg=True):
        """
        Create or retrieve OTFG from database
        :param otfg_entry: CASTEP styled OTFG entry.
        Can either be the name of library (e.g C9) or the full specification with element like:
        "O 2|1.1|15|18|20|20:21(qc=7)"

        The created OTFGData node will by default labelled by the fully entry.
        """

        in_db = cls.from_entry(otfg_entry)

        # No existing entry
        if len(in_db) == 0:
            instance = cls(otfg_entry=otfg_entry)
            if store_otfg:
                instance.store()

            # Automatically set the label
            instance.label = otfg_entry
            return (instance, True)

        # There is a existing identical enetry in the db
        else:
            if len(in_db) > 1:
                if use_first:
                    return (in_db[0], False)
                else:
                    pks = ", ".join([str(i.pk) for i in in_db])
                    raise ValueError(
                        "More than one duplicated OTFG data has been found. pks={}"
                        .format(pks))
            else:
                return (in_db[0], False)

    def set_string(self, otfg_entry):
        """Set the full string of OTFGData instance"""
        if self.element is None:
            self.set_element("LIBRARY")
        self.set_attribute("otfg_entry", str(otfg_entry))

    def set_element(self, element):
        """Set the element of OTFGData instance"""
        self.set_attribute("element", str(element))

    def store(self, *args, **kwargs):
        self._validate()
        return super(OTFGData, self).store(*args, **kwargs)

    @property
    def string(self):
        return self.get_attribute('otfg_entry', None)

    @property
    def element(self):
        """Element of the OTFG. May not be available"""
        return self.get_attribute('element', None)

    @property
    def entry(self):
        """Plain format of the OTFG"""
        string = self.string
        element = self.element
        if element is None or element == "LIBRARY":
            return string

        else:
            return element + " " + string

    @classmethod
    def from_entry(cls, entry):
        """
        Return a list of OTFG that matches with the string
        """

        from aiida.orm.querybuilder import QueryBuilder
        from .utils import split_otfg_entry

        element, string = split_otfg_entry(entry)
        qb = QueryBuilder()
        qb.append(cls,
                  filters={
                      'attributes.otfg_entry': {
                          '==': string
                      },
                      'attributes.element': {
                          '==': element
                      }
                  })

        return [i[0] for i in qb.all()]

    def _validate(self):
        """Validate the format of OTFG configuration"""
        super(OTFGData, self)._validate()
        if self.element is None:
            raise ValidationError("The value of element is not set. "
                                  "Set it to 'LIBRARY' manually to indicate. "
                                  "This is a library")

    @classproperty
    def otfg_family_type_string(cls):
        return OTFGGROUP_TYPE

    @classmethod
    def get_otfg_group(cls, group_label):
        """
        Return the OTFGData group with the given name.
        """
        return OTFGGroup.objects.get(label=group_label)

    @classmethod
    def get_otfg_groups(cls, filter_elements=None, user=None):
        """
        Return all names of groups of type otfg, possibly with some filters.

        :param filter_elements: A string or a list of strings.
               If present, returns only the groups that contains one Usp for
               every element present in the list. Default=None, meaning that
               all families are returned.
        :param user: if None (default), return the groups for all users.
               If defined, it should be either a DbUser instance, or a string
               for the user name (that is, the user email).
        """
        from aiida.orm import QueryBuilder
        from aiida.orm import User

        query = QueryBuilder()
        filters = {'type_string': {'==': cls.otfg_family_type_string}}

        query.append(OTFGGroup, filters=filters, tag='group', project='*')

        if user:
            query.append(User,
                         filters={'email': {
                             '==': user
                         }},
                         with_group='group')

        if isinstance(filter_elements, six.string_types):
            filter_elements = [filter_elements]

        if filter_elements is not None:
            actual_filter_elements = {_.capitalize() for _ in filter_elements}
            # LIBRARY is a wild card
            actual_filter_elements.add("LBIRARY")

            query.append(
                cls,
                filters={'attributes.element': {
                    'in': filter_elements
                }},
                with_group='group')

        query.order_by({'group': {'id': 'asc'}})
        return [_[0] for _ in query.all()]



"""
Module for generating text based CASTEP inputs
"""
from __future__ import absolute_import, print_function

import six
import numpy as np
from .datastructure import ParamFile, CellFile
from aiida.common import InputValidationError, MultipleObjectsError
from .utils import get_castep_ion_line, _lowercase_dict, _uppercase_dict

from aiida.orm import UpfData
from ..data.otfg import OTFGData
from ..data.usp import UspData
from six.moves import zip

from aiida_castep.common import INPUT_LINKNAMES as in_ln


class CastepInputGenerator(object):
    """
    Class for generating CASTEP inputs
    """
    def __init__(self):
        """
        Initialise the object
        """
        # Initialize the underlying cell file and param file
        # objects
        self.param_file = ParamFile()
        self.cell_file = CellFile()
        self.local_copy_list_to_append = set()

    def prepare_inputs(self, reset=True):
        """
        Prepare the inputs
        :param reset: Rest exisitng self.param_file and self.cell file
        """
        if reset:
            self.param_file = ParamFile()
            self.cell_file = CellFile()

        self.local_copy_list_to_append = set()
        param_dict = self.inputs[in_ln['parameters']].get_dict()
        settings_node = self.inputs.get('settings', None)
        settings_dict = settings_node.get_dict() if settings_node else {}

        # Standardise top level keys should be CAPITALIZED
        param_dict = _uppercase_dict(param_dict, dict_name="parameters")
        # Second level keys should be lowercased
        param_dict = {
            k: _lowercase_dict(v, dict_name=k)
            for k, v in six.iteritems(param_dict)
        }

        # Set iprint to 1
        param_dict["PARAM"]["iprint"] = param_dict["PARAM"].get("iprint", 1)

        # Set run_time using define value for this calcualtion
        run_time = self.inputs.metadata.options.get('max_wallclock_seconds')
        if run_time:
            n_seconds = run_time * 0.95
            n_seconds = (n_seconds //
                         60) * 60  # Round down to the nearest minutes
            # Do not do any thing if calculated time is less than 1 hour
            if n_seconds < 180:
                pass
            elif "run_time" not in param_dict["PARAM"]:
                param_dict["PARAM"]["run_time"] = int(n_seconds)

        # Set the default comment using the label of this calculation
        comment_str = self.inputs.metadata.get('label', None)
        if "comment" not in param_dict["PARAM"] and comment_str:
            param_dict["PARAM"]["comment"] = comment_str

        # Expose at the instance level
        self.param_dict = param_dict
        self.settings_dict = settings_dict

        # prepare the cell and param files
        self._prepare_cell_file()
        self._prepare_param_file()

    def _prepare_cell_file(self):
        """
        Prepare the cell file
        """

        cell_vector_list = []
        for vector in self.inputs[in_ln['structure']].cell:
            cell_vector_list.append(("{0:18.10f} {1:18.10f} "
                                     "{2:18.10f}".format(*vector)))

        self.cell_file["LATTICE_CART"] = cell_vector_list

        # --------- ATOMIC POSITIONS---------
        # for kind in self.inputs[in_ln['structure']].kinds:
        atomic_position_list = []
        mixture_count = 0
        # deal with initial spins
        spin_list = self.settings_dict.pop("SPINS", None)
        label_list = self.settings_dict.pop("LABELS", None)

        for i, site in enumerate(self.inputs[in_ln['structure']].sites):
            # get  the kind of the site
            kind = self.inputs[in_ln['structure']].get_kind(site.kind_name)

            # Position is always needed
            pos = site.position
            mixture = False
            try:
                name = kind.symbol
            # If we are dealing with mixed atoms
            except ValueError:
                name = kind.symbols
                mixture_count += 1
                mixture = True

            # If the symbol is not the same as the kindname
            # e.g there are inequivalent atoms of the same element
            # We change the name to '<symbol>:<kind.name>'
            if not mixture:
                # Only do this if the name(symbol) is not equal to the kindname
                if name != kind.name:
                    name = name + ':' + kind.name
            else:
                # If we are dealing with the mixtures,
                # we also add the kindname as an identifier
                name = [ntemp + ':' + kind.name for ntemp in name]

            if spin_list:
                spin = spin_list[i]
            else:
                spin = None

            # deal with labels
            if label_list:
                label = label_list[i]
            else:
                label = None

            # Get the line of positions_abs block
            line = get_castep_ion_line(name,
                                       pos,
                                       label=label,
                                       spin=spin,
                                       occupation=kind.weights,
                                       mix_num=mixture_count)

            # Append the line to the list
            atomic_position_list.append(line)

        # End of the atomic position block
        self.cell_file["POSITIONS_ABS"] = atomic_position_list

        # Check the consistency of spin in parameters
        if spin_list:
            total_spin = sum(s for s in spin_list if s)
            param_spin = self.param_dict["PARAM"].get("spin", None)
            if param_spin is not None:
                # If spin is specified - check consistency
                if param_spin != total_spin:
                    raise InputValidationError(
                        "Inconsistent spin in cell and param files."
                        "Total spin: {} in cell file but {} in param file".
                        format(total_spin, param_spin))
            else:
                # If no spin specified, do it automatically
                # Note that we don't check if spin polarized calculation is
                # requested in the first place
                self.param_dict["PARAM"]["spin"] = total_spin

        # --------- KPOINTS ---------
        kpoints = self.inputs.get('kpoints')
        use_kpoints = self.inputs.metadata.options.use_kpoints
        if not kpoints and use_kpoints:
            raise InputValidationError(
                'Kpoints required but not found in the input')

        if self.inputs.metadata.options.use_kpoints:
            try:
                mesh, offset = kpoints.get_kpoints_mesh()
                has_mesh = True

            except AttributeError:
                try:
                    kpoints_list = kpoints.get_kpoints()
                    num_kpoints = len(kpoints_list)
                    has_mesh = False
                    if num_kpoints == 0:
                        raise InputValidationError(
                            "At least one k points must be provided")
                except AttributeError:
                    raise InputValidationError(
                        "No valid kpoints have been found")

                try:
                    _, weights = kpoints.get_kpoints(also_weights=True)

                except AttributeError:
                    weights = np.ones(num_kpoints, dtype=float) / num_kpoints

            kpoints_line_list = []
            if has_mesh is True:
                self.cell_file["kpoints_mp_grid"] = "{} {} {}".format(*mesh)
                if offset != [0., 0., 0.]:
                    self.cell_file["kpoints_mp_offset"] = "{} {} {}".format(
                        *offset)
            else:
                for kpoint, weight in zip(kpoints_list, weights):
                    kpoints_line_list.append("{:18.10f} {:18.10f} "
                                             "{:18.10f} {:18.10f}".format(
                                                 kpoint[0], kpoint[1],
                                                 kpoint[2], weight))
                self.cell_file["KPOINTS_LIST"] = kpoints_line_list

        # --------- keywords in cell file---------
        for key, value in six.iteritems(self.param_dict["CELL"]):

            if "species_pot" in key:
                raise MultipleObjectsError(
                    "Pseudopotentials should not be specified directly")

            # TODO use the castepinput package
            # Constructing block keywrods
            # We identify the key should be treated as a block it
            # is not a string and has len() > 0
            self.cell_file[key] = value

        self._prepare_pseudo_potentials()

    def _include_extra_kpoints(self,
                               kpn_node,
                               kpn_name,
                               kpn_settings,
                               report_fn=None):
        """Write extra kpoints to the cell"""

        try:
            mesh, offset = kpn_node.get_kpoints_mesh()
            has_mesh = True
        except AttributeError:
            # Not defined as mesh
            try:
                bs_kpts_list = kpn_node.get_kpoints()
                num_kpoints = len(bs_kpts_list)
                has_mesh = False
                if num_kpoints == 0:
                    raise InputValidationError(
                        "At least one k points must be provided")
            except AttributeError:
                raise InputValidationError(
                    "No valid {}_kpoints have been found from node {}".format(
                        kpn_name.lower(), kpn_node.pk))

            # Do we have weights defined?
            try:
                _, weights = kpn_node.get_kpoints(also_weights=True)
            except AttributeError:
                # If not, fill with fractions
                if kpn_settings['need_weights'] is True:
                    import numpy as np
                    weights = np.ones(num_kpoints, dtype=float) / num_kpoints
                    if report_fn is not None:
                        report_fn(
                            'Warning:filling evenly distributed weights for {}_kpoints'
                            .format(kpn_name))

        # now add to the cell file
        if has_mesh is True:
            mesh_name = "{}_kpoint_mp_grid".format(kpn_name)
            self.cell_file[mesh_name] = "{} {} {}".format(*mesh)
            if offset != [0., 0., 0.]:
                self.cell_file[mesh_name.replace(
                    "grid", "offset")] = "{} {} {}".format(*offset)
        else:
            extra_kpts_lines = []
            for kpoint, weight in zip(bs_kpts_list, weights):
                if kpn_settings['need_weights'] is True:
                    extra_kpts_lines.append("{:18.10f} {:18.10f} "
                                            "{:18.10f} {:18.14f}".format(
                                                kpoint[0], kpoint[1],
                                                kpoint[2], weight))
                else:
                    extra_kpts_lines.append("{:18.10f} {:18.10f} "
                                            "{:18.10f}".format(
                                                kpoint[0], kpoint[1],
                                                kpoint[2]))
            bname = "{}_kpoint_list".format(kpn_name).upper()
            self.cell_file[bname] = extra_kpts_lines

    def _prepare_pseudo_potentials(self):

        # --------- PSEUDOPOTENTIALS --------
        # Check if we are using UPF pseudos
        # Now only support simple elemental pseudopotentials

        species_pot_map = {}
        pseudos = self.inputs.pseudos
        # Make kindname unique
        for kind in self.inputs[in_ln['structure']].kinds:
            symbols = kind.symbols
            # If the site has multiple symbols, add all of them to the list
            if len(symbols) > 1:
                mixture = True
            else:
                mixture = False
            for symbol in symbols:
                if symbol == kind.name:
                    pseudo_name = symbol
                else:
                    pseudo_name = symbol + ':' + kind.name

                if not mixture:
                    # Get the pseudopotential is defined by the kind.name
                    ps = pseudos[kind.name]
                else:
                    # If with mixture the pseudopotential is deined as '<kind_name>_<symbol>'
                    ps = pseudos[kind.name + '_' + symbol]

                # If we are dealing with a UpfData object
                if isinstance(ps, (UpfData, UspData)):
                    # Add the specification to the file
                    species_pot_map[pseudo_name] = "{:5} {}".format(
                        pseudo_name, ps.filename)
                    # Add to the copy list
                    self.local_copy_list_to_append.add(
                        (ps.uuid, ps.filename, ps.filename))
                # If we are using OTFG, just add the string property of it
                elif isinstance(ps, OTFGData):
                    species_pot_map[pseudo_name] = "{:5} {}".format(
                        pseudo_name, ps.string)
                else:
                    raise InputValidationError(
                        'Unkonwn node as pseudo: {}'.format(ps))

        # Ensure it is a list
        self.cell_file["SPECIES_POT"] = list(species_pot_map.values())

    def _prepare_param_file(self):
        """
        Prepare the content of PARAM file
        """
        self.param_file.update(self.param_dict["PARAM"])



"""
Utility module with useful functions
"""
import io
import numpy as np

# pylint: disable=import-outside-toplevel, too-many-locals


def atoms_to_castep(atoms, index):
    """Convert ase atoms' index to castep like
    return (Specie, Ion) Deprecated, use ase_to_castep_index"""
    atom = atoms[index]
    symbol = atom.symbol
    # Start counter
    count = 0
    for atom in atoms:
        if atom.symbol == symbol:
            count += 1
        if atom.index == index:
            break
    return symbol, count


def ase_to_castep_index(atoms, indices):
    """Convert a list of indices to castep style
    return list of (element, i in the same element)"""
    if isinstance(indices, int):
        indices = [indices]
    symbols = np.array(atoms.get_chemical_symbols())
    iatoms = np.arange(len(atoms))
    res = []
    # Iterate through given indices
    for i in indices:
        mask = symbols == symbols[i]  # Select the same species

        # CASTEP start counting from 1
        counter = np.where(iatoms[mask] == i)[0][0] + 1
        res.append([symbols[i], counter])
    return res


def generate_ionic_fix_cons(atoms, indices, mask=None, count_start=1):
    """
    create ionic constraint section via indices and ase Atoms

    :param atoms: atoms object to be fixed
    :param indices: indices of the atoms to be fixed
    :param mask: a list of 3 integers, must be 0 (no fix) or 1 (fix this Cartesian)

    :returns: (lines, index of the next constraint)

    """

    # Convert to castep style indices (element, index)
    castep_indices = ase_to_castep_index(atoms, indices)
    count = count_start
    lines = []
    if mask is None:
        mask = (1, 1, 1)
    for symbol, i in castep_indices:
        if mask[0]:
            lines.append("{:<4d} {:<2}    {:<4d} 1 0 0".format(
                count, symbol, i))
        if mask[1]:
            lines.append("{:<4d} {:<2}    {:<4d} 0 1 0".format(
                count + 1, symbol, i))
        if mask[2]:
            lines.append("{:<4d} {:<2}    {:<4d} 0 0 1".format(
                count + 2, symbol, i))
        count += sum(mask)
    return lines, count


def generate_rel_fix(atoms, indices, ref_index=0, count_start=1):
    """
    Generate relative constraints

    .. note::

      In CASTEP the mass of atoms are coupled in the fix,
      hence to truelly fix the relative positions you will have to
      declare all atoms having the same weight using the
      SPECIES_MASS block.

    :param atoms: atoms object to be fixed

    :param indices: indices of the atoms to be fixed

    :param ref_index: index of the reference atom in amoung the atoms
      being fixed. Default is the first atom appear in the indices.

    :returns: (lines, index of the next constraint)
    """
    castep_indices = ase_to_castep_index(atoms, indices)
    lines = []
    count = count_start
    symbol_ref, i_ref = castep_indices[ref_index]
    for symbol, i in castep_indices[1:]:
        lines.append("{:<4d} {:<2}    {:<4d} 1 0 0".format(count, symbol, i))
        lines.append("{:<4d} {:<2}    {:<4d} -1 0 0".format(
            count, symbol_ref, i_ref))
        lines.append("{:<4d} {:<2}    {:<4d} 0 1 0".format(
            count + 1, symbol, i))
        lines.append("{:<4d} {:<2}    {:<4d} 0 -1 0".format(
            count + 1, symbol_ref, i_ref))
        lines.append("{:<4d} {:<2}    {:<4d} 0 0 1".format(
            count + 2, symbol, i))
        lines.append("{:<4d} {:<2}    {:<4d} 0 0 -1".format(
            count + 2, symbol_ref, i_ref))

        count += 3
    return lines, count


def castep_to_atoms(atoms, specie, ion):
    """Convert castep like index to ase Atoms index"""
    return [atom for atom in atoms if atom.symbol == specie][ion - 1].index


def sort_atoms_castep(atoms, copy=True, order=None):
    """
    Sort ``ase.Atoms`` instance  to castep style.
    A sorted ``Atoms`` will have the same index before and after calculation.
    This is useful when chaining optimisation requires specifying per atoms
    tags such as *SPIN* and *ionic_constraints*.
    :param order: orders of coordinates. (0, 1, 2) means the sorted atoms
    will be ascending by x, then y, then z if there are equal x or ys.

    :returns: A ``ase.Atoms`` object that is sorted.
    """
    _ = copy
    # Sort castep style
    if order is not None:
        for i in reversed(order):
            isort = np.argsort(atoms.positions[:, i], kind="mergesort")
            atoms.positions = atoms.positions[isort]
            atoms.numbers = atoms.numbers[isort]

    isort = np.argsort(atoms.numbers, kind="mergesort")
    atoms = atoms[isort]
    return atoms


def desort_atoms_castep(atoms, original_atoms):
    """
    Recover the original ordering of the atoms. CASTEP sort the atoms
    in the order of atomic number and then the order of appearance internally.

    :param copy: If True then return an copy of the atoms
    :returns: A ``ase.Atoms`` object that has be sorted
    """

    isort = np.argsort(original_atoms.numbers, kind='mergesort')
    rsort = np.zeros(isort.shape, dtype=int)
    rsort.fill(-1)
    for i, j in enumerate(isort):
        rsort[j] = i

    return atoms[rsort]


def is_castep_sorted(atoms):
    """
    Check if an Atoms object is CASTEP style sorted
    """
    numbers = np.asarray(atoms.numbers)
    return np.all(numbers == np.sort(numbers))


def reuse_kpoints_grid(grid, lowest_pk=False):
    """
    Retrieve previously stored kpoints mesh data node.
    If there is no such ``KpointsData``, a new node will be created.
    Will return the one with highest pk
    :param grid: Grid to be retrieved
    :param bool lowest_pk: If set to True will return the node with lowest pk

    :returns: A KpointsData node representing the grid requested
    """
    from aiida.orm import QueryBuilder
    from aiida.orm import KpointsData
    qbd = QueryBuilder()
    qbd.append(KpointsData,
               tag="kpoints",
               filters={
                   "attributes.mesh.0": grid[0],
                   "attributes.mesh.1": grid[1],
                   "attributes.mesh.2": grid[2]
               })
    if lowest_pk:
        order = "asc"
    else:
        order = "desc"
    qbd.order_by({"kpoints": [{"id": {"order": order}}]})
    if qbd.count() >= 1:

        return qbd.first()[0]
    kpoints = KpointsData()
    kpoints.set_kpoints_mesh(grid)
    return kpoints


def traj_to_atoms(traj, combine_ancesters=False, eng_key="enthalpy"):
    """
    Generate a list of ASE Atoms given an AiiDA TrajectoryData object
    :param bool combine_ancesters: If true will try to combine trajectory
    from ancestor calculations

    :returns: A list of atoms for the trajectory.
    """
    from ase import Atoms
    from ase.calculators.singlepoint import SinglePointCalculator
    from aiida.orm import QueryBuilder, Node, CalcJobNode
    from aiida_castep.common import OUTPUT_LINKNAMES

    # If a CalcJobNode is passed, select its output trajectory
    if isinstance(traj, CalcJobNode):
        traj = traj.outputs.__getattr__(OUTPUT_LINKNAMES['trajectory'])
    # Combine trajectory from ancesters
    if combine_ancesters is True:
        qbd = QueryBuilder()
        qbd.append(Node, filters={"uuid": traj.uuid})
        qbd.append(CalcJobNode, tag="ans", ancestor_of=Node)
        qbd.order_by({"ans": "id"})
        calcs = [_[0] for _ in qbd.iterall()]
        atoms_list = []
        for counter in calcs:
            atoms_list.extend(
                traj_to_atoms(counter.outputs.__getattr__(
                    OUTPUT_LINKNAMES['trajectory']),
                              combine_ancesters=False,
                              eng_key=eng_key))
        return atoms_list
    forces = traj.get_array("forces")
    symbols = traj.get_array("symbols")
    positions = traj.get_array("positions")
    try:
        eng = traj.get_array(eng_key)
    except KeyError:
        eng = None
    cells = traj.get_array("cells")
    atoms_traj = []
    for counter, pos, eng_, force in zip(cells, positions, eng, forces):
        atoms = Atoms(symbols=symbols, cell=counter, pbc=True, positions=pos)
        calc = SinglePointCalculator(atoms, energy=eng_, forces=force)
        atoms.set_calculator(calc)
        atoms_traj.append(atoms)
    return atoms_traj


def get_remote_folder_info(calc, transport):
    """Get the information of the remote folder of a calculation"""
    path = calc.outputs.remote_folder.get_remote_path()
    transport.chdir(path)
    lsattrs = transport.listdir_withattributes()
    return lsattrs


def get_remote_folder_size(calc, transport):
    """
    Return size of a remote folder of a calculation in MB
    """
    try:
        lsattrs = get_remote_folder_info(calc, transport)
    except IOError:
        return 0
    total_size = 0
    for attr in lsattrs:
        total_size += attr['attributes'].st_size  # in byres
    return total_size / (1024)**2


def take_popn(seed):
    """
    Take section of population analysis from a seed.castep file
    Return a list of StringIO of the population analysis section
    """
    popns = []
    rec = False
    with open(seed + '.castep') as fhd:
        for line in fhd:
            if "Atomic Populations (Mulliken)" in line:
                record = io.StringIO()
                rec = True

            # record information
            if rec is True:
                if line.strip() == "":
                    rec = False
                    record.seek(0)
                    popns.append(record)
                else:
                    record.write(line)

    return popns


def read_popn(fname):
    """Read population file into pandas dataframe"""
    import pandas as pd
    table = pd.read_table(fname,
                          sep=r"\s\s+",
                          header=2,
                          comment="=",
                          engine="python")
    return table


def export_calculation(node, output_dir, prefix=None):
    """
    Export one calculation a a directory
    """
    from aiida.orm.utils.repository import FileType
    from pathlib import Path

    output_dir = Path(output_dir)
    output_dir.mkdir(exist_ok=True)

    def bwrite(node, outpath):
        """Write the objects stored under a node to a certain path"""
        for objname in node.list_object_names():
            if node.get_object(objname).type != FileType.FILE:
                continue
            with node.open(objname, mode='rb') as fsource:
                name, suffix = objname.split('.')
                if prefix and name == node.get_option('seedname'):
                    outname = prefix + '.' + suffix
                else:
                    outname = objname
                fpath = str(outpath / outname)
                with open(fpath, 'wb') as fout:
                    readlength = 1024 * 512  # 1MB
                    while True:
                        buf = fsource.read(readlength)
                        if buf:
                            fout.write(buf)
                        else:
                            break

    #inputs
    bwrite(node, output_dir)

    # outputs
    retrieved = node.outputs.retrieved
    bwrite(retrieved, output_dir)


def compute_kpoints_spacing(cell, grid, unit="2pi"):
    """
    Compute the spacing of the kpoints in the reciprocal space.
    Spacing = 1 / cell_length / mesh for each dimension.
    Assume orthogonal cell shape.
    """
    cell = np.asarray(cell, dtype=np.float)
    grid = np.asarray(grid, dtype=np.float)

    spacings = 1. / cell / grid
    if unit == "1/A":
        spacings *= 2 * np.pi
    elif unit != "2pi":
        raise ValueError("Unit {} is not unkown".format(unit))
    return spacings



"""
Fixture for reusing existing test db
"""

from __future__ import absolute_import
import os

from aiida.backends.testbase import check_if_tests_can_run
from aiida.manage.fixtures import FixtureManager
from contextlib import contextmanager


class ExistingProfileFixtureManager(FixtureManager):
    """
    A fixture manager that reuses the existing profile
    """
    def __init__(self):
        super(ExistingProfileFixtureManager, self).__init__()
        self._is_running_on_test_profile = False

    def create_profile(self):
        """Not actually create the profile but reuse the test profile"""
        test_profile = os.environ.get('AIIDA_TEST_PROFILE', None)
        if not test_profile:
            return super(ExistingProfileFixtureManager, self).create_profile()

        from aiida import load_dbenv
        load_dbenv(profile=test_profile)
        check_if_tests_can_run()

        # Running on test profile
        self._is_running_on_test_profile = True

        from aiida.backends.djsite.db.testbase import DjangoTests
        self._test_case = DjangoTests()
        # Load the dbenv

    def destroy_all(self):
        """Do not destroy anything"""
        pass

    def has_profile_open(self):
        return self._is_running_on_test_profile or \
            super(ExistingProfileFixtureManager, self).has_profile_open()


_GLOBAL_FIXTURE_MANAGER = ExistingProfileFixtureManager()


@contextmanager
def fixture_manager():
    try:
        if not _GLOBAL_FIXTURE_MANAGER.has_profile_open():
            _GLOBAL_FIXTURE_MANAGER.backend = 'django'
            _GLOBAL_FIXTURE_MANAGER.create_profile()
        yield _GLOBAL_FIXTURE_MANAGER
    finally:
        _GLOBAL_FIXTURE_MANAGER.reset_db()



"""
Tools for calculations
"""
from __future__ import absolute_import
from __future__ import print_function
import six
import warnings

from aiida.tools import CalculationTools
from aiida.common import InputValidationError
from aiida.orm import CalcJobNode, Dict
from aiida.common.links import LinkType
from aiida.plugins import DataFactory
from aiida.engine import CalcJob, ProcessBuilder

from aiida_castep.common import INPUT_LINKNAMES, OUTPUT_LINKNAMES
from six.moves import zip

__all__ = [
    'CastepCalcTools', 'create_restart', 'castep_input_summary',
    'update_parameters', 'use_pseudos_from_family'
]


class CastepCalcTools(CalculationTools):
    def get_castep_input_summary(self):
        return castep_input_summary(self._node)

    def compare_with(self, the_other_calc, reverse=False):
        """
        Compare with another calculation
        Look for difference in get_castep_input_summary functions
        :params node: pk or uuid or node
        :params reverse: reverse the comparison, by default this node
        is the "new" and the one compared with is "old".
        """
        if isinstance(the_other_calc, (int, six.string_types)):
            from aiida.orm import load_node
            calc2 = load_node(the_other_calc)
        else:
            calc2 = the_other_calc

        from deepdiff import DeepDiff
        this_param = castep_input_summary(self._node)
        other_param = castep_input_summary(calc2)
        if reverse is True:
            res = DeepDiff(this_param, other_param)
        else:
            res = DeepDiff(other_param, this_param)

        return res

    def create_restart(self,
                       ignore_state=False,
                       restart_mode='restart',
                       use_output_structure=False,
                       **kwargs):
        if self._node.exit_status != 0 and not ignore_state:
            raise RuntimeError(
                'exit_status is not 0. Set ignore_state to ignore')

        builder = create_restart(self._node.get_builder_restart(),
                                 calcjob=self._node,
                                 restart_mode=restart_mode,
                                 **kwargs)

        # Carry over the label
        builder.metadata.label = self._node.label

        if use_output_structure is True:
            builder[
                INPUT_LINKNAMES['structure']] = self._node.outputs.__getattr__(
                    OUTPUT_LINKNAMES['structure'])

        if restart_mode == 'continuation' or kwargs.get('reuse'):
            builder[INPUT_LINKNAMES[
                'parent_calc_folder']] = self._node.outputs.__getattr__(
                    'remote_folder')

        return builder


def use_pseudos_from_family(builder, family_name):
    """
    Set the pseudos port namespace for a builder using pseudo family name
    :note: The structure must already be set in the builder.

    :param builder: ProcessBuilder instance to be processed, it must have a structure
    :param family_name: the name of the group containing the pseudos
    :returns: The same builder with the pseudopotential set
    """
    from collections import defaultdict
    from aiida_castep.data import get_pseudos_from_structure

    # A dict {kind_name: pseudo_object}
    # But we want to run with use_pseudo(pseudo, kinds)

    structure = builder.get(INPUT_LINKNAMES['structure'], None)
    if structure is None:
        raise RuntimeError('The builder must have a StructureData')
    kind_pseudo_dict = get_pseudos_from_structure(structure, family_name)
    for kind, pseudo in kind_pseudo_dict.items():
        builder.pseudos.__setattr__(kind, pseudo)
    return builder


def castep_input_summary(calc):
    """
    Convenient fuction for getting a summary of the
    input of this calculation

    :param calc: A CalcJobNode or ProcessBuilder or a nested input dictionary
    :returns: A dictionary
    """

    out_info = {}
    # Check what is passed
    if isinstance(calc, CalcJobNode):
        inp_dict = calc.get_incoming(link_type=(LinkType.INPUT_CALC,
                                                LinkType.INPUT_WORK)).nested()
        options = calc.get_options()
        metadata = {}  # Metadata is empty when Node is passed
        is_node = True
    elif isinstance(calc, ProcessBuilder):
        # Case of builder
        inp_dict = calc._data
        metadata = calc.metadata._data
        options = calc.metadata.get('options', {})
        is_node = False
    elif isinstance(calc, dict):
        # Case of a input dictionary
        inp_dict = calc
        metadata = calc.get('metadata', {})
        options = metadata.get('options', {})
        is_node = False

    def get_node(label):
        """Get node from input dictionary"""
        return inp_dict.get(INPUT_LINKNAMES[label])

    in_param = get_node('parameters')
    in_kpn = get_node('kpoints')
    in_settings = get_node('settings')
    in_structure = get_node('structure')
    in_code = inp_dict.get('code')
    in_remote = get_node('parent_calc_folder')
    pseudos = inp_dict.get('pseudos')

    param_dict = in_param.get_dict()
    out_info.update(param_dict)

    out_info["kpoints"] = in_kpn.get_description()
    out_info["structure"] = {
        "formula": in_structure.get_formula(),
        "cell": in_structure.cell,
        "label": in_structure.label
    }

    out_info["code"] = in_code
    out_info["computer"] = calc.computer if is_node else in_code.computer
    out_info["resources"] = options.get('resources')
    out_info["custom_scheduler_commands"] = options.get(
        'custom_scheduler_commands')
    out_info["qos"] = options.get('qos')
    out_info["account"] = options.get('account')
    out_info["wallclock"] = options.get('max_wallclock_seconds')
    out_info["label"] = calc.label if is_node else metadata.get('label')
    out_info["description"] = calc.description if is_node else metadata.get(
        'description')

    # Show the parent calculation whose RemoteData is linked to the node
    if in_remote is not None:
        input_calc = [
            n.node for n in in_remote.get_incoming(link_type=LinkType.CREATE)
        ]
        assert len(
            input_calc
        ) < 2, "More than one JobCalculation found, something seriously wrong"
        if input_calc:
            input_calc = input_calc[0]
            out_info["parent_calc"] = {
                "pk": input_calc.pk,
                "label": input_calc.label
            }
        out_info["parent_calc_folder"] = in_remote

    if in_settings is not None:
        out_info["settings"] = in_settings.get_dict()
    out_info["pseudos"] = pseudos
    return out_info


def update_parameters(inputs, force=False, delete=None, **kwargs):
    """
    Convenient function to update the parameters of the calculation.
    Will atomiatically set the PARAM or CELL field in unstored
    ParaemterData linked to the calculation.
    If no ``Dict`` is linked to the calculation, a new node will be
    created.

    ..note:
      This method relies on the help information to check and assign
      keywords to PARAM or CELL field of the Dict
      (i.e for generating .param and .cell file)

    calc.update_parameters(task="singlepoint")

    :param force: flag to force the update even if the Dict node is stored.
    :param delete: A list of the keywords to be deleted.
    """
    param_node = inputs.get(INPUT_LINKNAMES['parameters'])

    # Create the node if none is found
    if param_node is None:
        warnings.warn("No existing Dict node found, creating a new one.")
        param_node = Dict(dict={"CELL": {}, "PARAM": {}})
        inputs[INPUT_LINKNAMES['parameters']] = param_node

    if isinstance(param_node, Dict) and param_node.is_stored:
        if force:
            # Create a new node if the existing node is stored
            param_node = Dict(dict=param_node.get_dict())
            inputs[INPUT_LINKNAMES['parameters']] = param_node
        else:
            raise RuntimeError("The input Dict<{}> is already stored".format(
                param_node.pk))

    # If the `node` is just a plain dict, we keep it that way
    if isinstance(param_node, Dict):
        param_dict = param_node.get_dict()
        py_dict = False
    else:
        param_dict = param_node
        py_dict = True

    # Update the dictionary
    from .helper import HelperCheckError, CastepHelper
    helper = CastepHelper()
    dict_update, not_found = helper._from_flat_dict(kwargs)
    if not_found:
        suggest = [helper.get_suggestion(i) for i in not_found]
        error_string = "Following keys are invalid -- "
        for error_key, sug in zip(not_found, suggest):
            error_string += "{}: {}; ".format(error_key, sug)
        raise HelperCheckError(error_string)
    else:
        param_dict["PARAM"].update(dict_update["PARAM"])
        param_dict["CELL"].update(dict_update["CELL"])

    # Delete any keys as requested
    if delete:
        for key in delete:
            tmp1 = param_dict["PARAM"].pop(key, None)
            tmp2 = param_dict["CELL"].pop(key, None)
            if (tmp1 is None) and (tmp2 is None):
                warnings.warn("Key '{}' not found".format(key))

    # Apply the change to the node
    if py_dict:
        inputs[INPUT_LINKNAMES['parameters']] = param_dict
    else:
        param_node.set_dict(param_dict)
    return inputs


def create_restart(inputs,
                   entry_point='castep.castep',
                   calcjob=None,
                   param_update=None,
                   param_delete=None,
                   restart_mode='restart',
                   use_castep_bin=False,
                   parent_folder=None,
                   reuse=False):
    """
    Function to create a restart for a calculation.
    :param inputs: A builder or nested dictionary
    :param entry_point: Name of the entry points
    :param param_update: Update the parameters
    :param param_delete: A list of parameters to be deleted
    :param restart_mode: Mode of the restart, 'continuation' or 'restart'
    :param use_castep_bin: Use hte 'castep_bin' file instead of check
    :param parent_folder: Remote folder to be used for restart
    :param reuse: Use the reuse mode
    """
    from aiida.plugins import CalculationFactory
    from aiida.engine import ProcessBuilder

    # Create the builder, in any case
    if isinstance(inputs, dict):
        processclass = CalculationFactory(entry_point)
        builder = processclass.get_builder()
    elif isinstance(inputs, ProcessBuilder):
        builder = inputs._process_class.get_builder()

    builder._update(inputs)

    # Update list
    update = {}
    delete = []

    # Set the restart tag
    suffix = '.check' if not use_castep_bin else '.castep_bin'
    if restart_mode == 'continuation':
        update['continuation'] = 'parent/' + builder.metadata.seedname + suffix
        delete.append('reuse')
    elif restart_mode == 'restart' and reuse:
        update['reuse'] = 'parent/' + builder.metadata.seedname + suffix
        delete.append('continuation')
    elif restart_mode is None:
        delete.extend(['continuation', 'reuse'])
    elif restart_mode != 'restart':
        raise RuntimeError('Unknown restart mode: ' + restart_mode)

    if param_update:
        update.update(param_update)
    if param_delete:
        delete.extend(param_delete)

    new_builder = update_parameters(builder,
                                    force=True,
                                    delete=delete,
                                    **update)

    # Set the parent folder
    if parent_folder is not None:
        new_builder[INPUT_LINKNAMES['parent_calc_folder']] = parent_folder

    return new_builder


def validate_input_param(input_dict, allow_flat=False):
    """
    Validate inputs parameters
    :param input_dict: A Dict instance or python dict instance
    """

    from .helper import CastepHelper
    if isinstance(input_dict, Dict):
        py_dict = input_dict.get_dict()
    else:
        py_dict = input_dict
    helper = CastepHelper()
    helper.check_dict(py_dict, auto_fix=False, allow_flat=allow_flat)


def input_param_validator(input_dict, port=None):
    """
    Validator used for input ports
    """

    from .helper import HelperCheckError
    try:
        validate_input_param(input_dict)
    except HelperCheckError as error:
        return error.args[0]


def flat_input_param_validator(input_dict, port=None):
    """
    Validator that allows allow_flat parameter format
    """
    from .helper import HelperCheckError
    try:
        validate_input_param(input_dict, allow_flat=True)
    except HelperCheckError as error:
        return error.args[0]


def check_restart(builder, verbose=False):
    """
    Check the RemoteData reference by the builder is satisfied
    :returns: True if OK
    :raises: InputValidationError if error is found
    """
    import os
    from .utils import _lowercase_dict

    def _print(inp):
        if verbose:
            print(inp)

    paramdict = builder[INPUT_LINKNAMES['parameters']].get_dict()['PARAM']
    paramdict = _lowercase_dict(paramdict, "paramdict")
    stemp = paramdict.get("reuse", None)
    if not stemp:
        stemp = paramdict.get("continuation", None)
    if stemp is not None:
        fname = os.path.split(stemp)[-1]
        _print("This calculation requires a restart file: '{}'".format(fname))
    else:
        # No restart file needed
        _print("This calculation does not require a restart file.")
        return True

    # Now check if the remote folder has this file
    remote_data = builder.get(INPUT_LINKNAMES["parent_calc_folder"])
    if not remote_data:
        raise InputValidationError(
            "Restart requires "
            "parent_folder to be specified".format(fname))
    else:
        _print("Checking remote directory")
        folder_list = remote_data.listdir()
        if fname not in folder_list:
            raise InputValidationError(
                "Restart file {}"
                " is not in the remote folder".format(fname))
        else:
            _print("Check finished, restart file '{}' exists.".format(fname))
            return True



"""
conftest that prepares fixtures with tests involving orm of aiida
"""
from __future__ import absolute_import
import tempfile
import shutil
import pytest
import os
from aiida.common.exceptions import NotExistent
from aiida.common import AttributeDict
from pathlib import Path

this_folder = Path(__file__).parent


def get_backend_str():
    """ Return database backend string.

    Reads from 'TEST_AIIDA_BACKEND' environment variable.
    Defaults to django backend.
    """
    from aiida.backends.profile import BACKEND_DJANGO, BACKEND_SQLA
    backend_env = os.environ.get('TEST_AIIDA_BACKEND')
    if not backend_env:
        return BACKEND_DJANGO
    elif backend_env in (BACKEND_DJANGO, BACKEND_SQLA):
        return backend_env

    raise ValueError(
        "Unknown backend '{}' read from TEST_AIIDA_BACKEND environment variable"
        .format(backend_env))


@pytest.fixture(scope='session')
def aiida_profile():
    """setup a test profile for the duration of the tests
    If the environmental variable AIIDA_TEST_PROFILE is present
    will use an alternative fixture_manager that uses the test profile"""
    # import os
    # test_profile = os.environ.get('AIIDA_TEST_PROFILE', None)
    # if test_profile is not None:
    #     from fixture import fixture_manager
    # else:
    #     from aiida.manage.fixtures import fixture_manager
    from aiida.manage.fixtures import fixture_manager
    #from .fixture import fixture_manager
    with fixture_manager() as fixture_mgr:
        yield fixture_mgr


@pytest.fixture(scope='function')
def new_database(aiida_profile):
    """clear the database after each test"""
    yield aiida_profile
    aiida_profile.reset_db()


@pytest.fixture(scope="module")
def imps(aiida_profile):
    return Imports()


class Imports:
    def __init__(self):
        from aiida.plugins import CalculationFactory
        from aiida.plugins import DataFactory
        import aiida_castep.data.otfg as otfg
        from aiida_castep.data.otfg import OTFGData
        from aiida.orm import (KpointsData, StructureData, BandsData, Code,
                               Dict, Computer)
        for k, v in locals().items():
            if k not in ['self']:
                setattr(self, k, v)


class CastepTestApp(object):
    """
    A collection of methods to
    control the test
    """
    def __init__(self, profile, workdir):
        self._profile = profile
        self._workdir = Path(workdir)
        self.imps = Imports()

    def create_computer(self, **kwargs):
        """
        Return a generator for the Computers
        """
        from aiida.orm import Computer
        defaults = dict(name='localhost',
                        hostname='localhost',
                        transport_type='local',
                        scheduler_type='direct',
                        workdir=str(self._workdir))

        kwargs.update(defaults)
        computer = Computer(**kwargs).store()
        # Need to configure the computer before using
        # Otherwise there is no AuthInfo
        computer.configure()
        return computer

    def get_or_create_computer(self, **kwargs):
        """
        Fixture for a local computer called localhost.
        This is currently not in the AiiDA fixtures."""
        try:
            computer = self.imps.Computer.objects.get(**kwargs)
        except NotExistent:
            computer = self.create_computer(**kwargs)
        return computer

    @property
    def localhost(self):
        return self.get_or_create_computer(name='localhost')

    def code_mock_factory(self, overide=None):
        """Mock calculation, can overide by prepend path"""
        code = self.imps.Code()
        exec_path = this_folder.parent.parent / 'utils/mock.py'
        code.set_remote_computer_exec((self.localhost, str(exec_path)))
        code.set_input_plugin_name('castep.castep')
        if overide:
            code.set_prepend_text('export MOCK_CALC={}'.format(overide))
        return code

    @property
    def code_h2_geom(self):
        """A Code that always return H2-geom"""
        code = self.code_mock_factory('H2-geom')
        code.store()
        return code

    @property
    def code_echo(self):
        """Fixture of a code that just echos"""
        from aiida.orm import Code
        code = Code()
        code.set_remote_computer_exec((self.localhost, "/bin/echo"))
        code.set_input_plugin_name("castep.castep")
        code.store()
        return code

    @property
    def remotedata(self):
        """Create an remote data"""
        from aiida.orm import RemoteData
        rmd = RemoteData()
        rmd.computer = self.localhost
        rmd.set_remote_path(str(self._workdir))
        return rmd

    def get_kpoints_mesh(self, mesh):
        """Factory for kpoints with mesh"""
        kpoints_data = self.imps.KpointsData()
        kpoints_data.set_kpoints_mesh(mesh)
        return kpoints_data

    def upload_otfg_family(self, entries, name, desc='test', **kwargs):
        """Return a factory for upload OTFGS"""
        from aiida_castep.data.otfg import upload_otfg_family
        upload_otfg_family(entries, name, desc, **kwargs)

    @property
    def c9_otfg(self):
        return self.imps.OTFGData.get_or_create('C9')[0]

    @property
    def create_group(self):
        from aiida_castep.data.otfg import upload_otfg_family
        upload_otfg_family

    def get_builder(self):
        from aiida_castep.calculations.castep import CastepCalculation
        return CastepCalculation.get_builder()


@pytest.fixture
def create_otfg_group(db_test_app):
    def _create(otfgs, group_name):
        from aiida_castep.data.otfg import upload_otfg_family
        upload_otfg_family(otfgs, group_name, 'TEST', stop_if_existing=False)

    return _create


@pytest.fixture
def db_test_app(aiida_profile):
    """
    Yield an test app for controlling
    """

    workdir = tempfile.mkdtemp()
    app = CastepTestApp(aiida_profile, workdir)
    yield app
    aiida_profile.reset_db()
    shutil.rmtree(workdir)


@pytest.fixture
def inputs_default():
    options = {
        'seedname': 'aiida',
        'input_filename': 'aiida.cell',
        'output_filename': 'aiida.castep',
        'symlink_usage': True,
        'parent_folder_name': 'parent',
        'retrieve_list': [],
        'use_kpoints': True,
        'resources': {
            'num_machines': 1,
            'tot_num_mpiprocs': 1
        }
    }
    inputs = AttributeDict()
    inputs['metadata'] = AttributeDict()
    inputs.metadata['options'] = AttributeDict(options)
    return inputs


@pytest.fixture
def sto_calc_inputs(
        db_test_app,
        inputs_default,
):

    from ..utils import get_sto_structure
    sto_structure = get_sto_structure()
    inputs = inputs_default

    pdict = {
        "PARAM": {
            "task": "singlepoint"
        },
        "CELL": {
            "symmetry_generate": True,
            "cell_constraints": ['0 0 0', '0 0 0']
        }
    }
    # pdict["CELL"].pop("block species_pot")
    inputs.parameters = db_test_app.imps.Dict(dict=pdict)
    inputs.structure = sto_structure
    c9 = db_test_app.c9_otfg
    inputs.pseudos = AttributeDict({"Sr": c9, 'Ti': c9, 'O': c9})
    inputs.kpoints = db_test_app.get_kpoints_mesh((3, 3, 3))
    inputs.code = db_test_app.code_echo
    return inputs


@pytest.fixture
def sto_spectral_inputs(sto_calc_inputs, db_test_app):
    kpoints = db_test_app.imps.KpointsData()
    kpoints.set_kpoints([[0.0, 0.5, 0.5], [0.0, 0.0, 0.0]])
    sto_calc_inputs.spectral_kpoints = kpoints
    return sto_calc_inputs


@pytest.fixture
def sto_calc_builder(sto_calc_inputs):
    return inps_or_builder(sto_calc_inputs, 1)


@pytest.fixture(params=[0, 1])
def sto_calc_inps_or_builder(request, sto_calc_inputs):
    """Fixture that returns raw input dictionary or builder"""
    return inps_or_builder(sto_calc_inputs, request.param)


def inps_or_builder(inps, num):
    """Helper function to convert inputs to builder or do nothing"""
    if num == 0:
        return inps
    elif num == 1:
        from aiida_castep.calculations.castep import CastepCalculation
        builder = CastepCalculation.get_builder()
        builder._update(inps)
        return builder
    else:
        raise RuntimeError('Not implemented')


@pytest.fixture
def h2_calc_inputs(
        inputs_default,
        db_test_app,
        sto_calc_inputs,
        h2_structure,
):

    inputs = inputs_default
    pdict = {
        "PARAM": {
            "task": "geometryoptimisation"
        },
        "CELL": {
            "symmetry_generate": True,
            "cell_constraints": ['0 0 0', '0 0 0']
        }
    }
    # pdict["CELL"].pop("block species_pot")
    inputs.parameters = db_test_app.imps.Dict(dict=pdict)
    inputs.structure = h2_structure
    inputs.pseudos = AttributeDict({"H": db_test_app.c9_otfg})
    inputs.kpoints = db_test_app.get_kpoints_mesh((3, 3, 3))
    inputs.code = db_test_app.code_h2_geom

    return inputs


@pytest.fixture
def h2_structure(aiida_profile, db_test_app):
    StructureData = db_test_app.imps.DataFactory("structure")
    a = 10

    cell = ((a, 0., 0.), (0., a, 0.), (0., 0., a))
    s = StructureData(cell=cell)
    s.append_atom(position=(0., 0., 0.), symbols=["H"])
    s.append_atom(position=(a / 2, a / 2, a / 2), symbols=["H"])
    s.label = "h2"
    return s


def test_localhost_fixture(db_test_app):
    """
    Test the localhost fixture
    """
    localhost = db_test_app.localhost
    localhost.name == "localhost"
    assert localhost.pk is not None


def test_code_fixture(db_test_app):
    """
    Test the localhost fixture
    """
    code_echo = db_test_app.code_echo
    assert code_echo.uuid is not None
    code_echo.get_remote_exec_path()


def test_remotedata_fixture(db_test_app):
    assert db_test_app.remotedata
    assert db_test_app.remotedata.get_remote_path() == str(
        db_test_app._workdir)


@pytest.fixture
def generate_calc_job_node(db_test_app):
    """
    Generate CalcJobNode
    """
    from aiida.orm import Node

    def _generate_calc_job_node(
            entry_point_name,
            results_folder,
            inputs=None,
            computer=None,
            outputs=None,
    ):
        """
        Generate a CalcJob node with fake retrieved node in the
        tests/data
        """
        from aiida.common.links import LinkType
        from aiida.orm import CalcJobNode, FolderData
        from aiida.plugins import CalculationFactory
        from aiida.plugins.entry_point import format_entry_point_string

        calc_class = CalculationFactory(entry_point_name)
        entry_point = format_entry_point_string('aiida.calculations',
                                                entry_point_name)
        builder = calc_class.get_builder()

        if not computer:
            computer = db_test_app.localhost
        node = CalcJobNode(computer=computer, process_type=entry_point)

        # Monkypatch the inputs
        if inputs is not None:
            inputs = AttributeDict(inputs)
            node.__dict__['inputs'] = inputs
            # Add direct inputs, pseudos are omitted
            for k, v in inputs.items():
                if isinstance(v, Node):
                    if not v.is_stored:
                        v.store()
                    node.add_incoming(v,
                                      link_type=LinkType.INPUT_CALC,
                                      link_label=k)

        options = builder.metadata.options
        node.set_attribute('input_filename', options.input_filename)
        node.set_attribute('seedname', options.seedname)
        node.set_attribute('output_filename', options.output_filename)
        node.set_attribute('error_filename', 'aiida.err')
        node.set_option('resources', {
            'num_machines': 1,
            'num_mpiprocs_per_machine': 1
        })
        node.set_option('max_wallclock_seconds', 1800)
        node.store()

        filepath = this_folder.parent / 'data' / results_folder
        retrieved = FolderData()
        retrieved.put_object_from_tree(str(filepath.resolve()))
        retrieved.add_incoming(node,
                               link_type=LinkType.CREATE,
                               link_label='retrieved')
        retrieved.store()

        if outputs is not None:
            for label, out_node in outputs.items():
                out_node.add_incoming(node,
                                      link_type=LinkType.CREATE,
                                      link_label=label)
                if not out_node.is_stored:
                    out_node.store()

        return node

    return _generate_calc_job_node


@pytest.fixture
def generate_parser():
    def _generate_parser(entry_point_name):
        from aiida.plugins import ParserFactory
        return ParserFactory(entry_point_name)

    return _generate_parser



"""
Module for parsing .castep file
This module should not rely on any of AiiDA modules
"""

from __future__ import absolute_import
import re
import logging
from collections import defaultdict
from six.moves import map
from six.moves import range

import numpy as np

from aiida_castep.parsers.utils import CASTEPOutputParsingError
from aiida_castep.common import EXIT_CODES_SPEC
from .._version import CALC_PARSER_VERSION

# pylint: disable=invalid-name,logging-format-interpolation,too-many-instance-attributes,too-many-branches,too-many-statements,too-many-locals
LOGGER = logging.getLogger("aiida")

__version__ = CALC_PARSER_VERSION

# NOTE one day we should allow the CODATA sets to be selected

# CODATA1986 (included herein for the sake of completeness)
# taken from
#    http://physics.nist.gov/cuu/Archive/1986RMP.pdf
units_CODATA1986 = {
    'hbar': 6.5821220E-16,  # eVs
    'Eh': 27.2113961,  # eV
    'kB': 8.617385E-5,  # eV/K
    'a0': 0.529177249,  # A
    'c': 299792458,  # m/s
    'e': 1.60217733E-19,  # C
    'me': 5.485799110E-4
}  # u

# CODATA2002: default in CASTEP 5.01
# (-> check in more recent CASTEP in case of numerical discrepancies?!)
# taken from
#    http://physics.nist.gov/cuu/Document/all_2002.pdf

units_CODATA2002 = {
    'hbar': 6.58211915E-16,  # eVs
    'Eh': 27.2113845,  # eV
    'kB': 8.617343E-5,  # eV/K
    'a0': 0.5291772108,  # A
    'c': 299792458,  # m/s
    'e': 1.60217653E-19,  # C
    'me': 5.4857990945E-4
}  # u

units_CODATA2010 = {
    'c': 299792458,
    'hbar': 6.58211928e-16,  # eV
    'e': 1.602176565e-19,  # C
    'Av': 6.02214129e23,  # Avogadro constant
    'R': 8.3144621,  # Gas constrant
    'Eh': 27.21138505,  # eV
    'a0': 0.52917721092,  # A - bohr radius
    'me': 5.4857990946e-4,  # u
    'kB': 8.6173324E-5  # eV/K
}

# (common) derived entries
for d in (units_CODATA1986, units_CODATA2002, units_CODATA2010):
    d['t0'] = d['hbar'] / d['Eh']  # s
    d['Pascal'] = d['e'] * 1E30  # Pa
units = units_CODATA2010

unit_suffix = "_units"

SCF_FAILURE_ERROR = "ERROR_SCF_NOT_CONVERGED"
GEOM_FAILURE_MESSAGE = "Maximum geometry optimization cycle has been reached"
END_NOT_FOUND_ERROR = "ERROR_NO_END_OF_CALCULATION"
INSUFFICENT_TIME_ERROR = "ERROR_TIMELIMIT_REACHED"
STOP_REQUESTED_ERROR = "ERROR_STOP_REQUESTED"


class RawParser:
    """An raw parser object to parse the output of CASTEP"""
    def __init__(self, out_lines, input_dict, md_geom_info, bands_lines,
                 **parser_opts):
        """Instantiate the parser by passing list of the lines"""

        self.dot_castep_lines = out_lines
        self.input_dict = input_dict
        self.md_geom_info = md_geom_info
        self.bands_lines = bands_lines
        self.parser_opts = parser_opts

        self.bands_res = {}
        self.traj_data = {}
        self.warnings = []
        self.geom_data = {}
        self.dot_castep_critical_message = []
        self.dot_castep_traj = {}
        self.dot_castep_data = {}

    def parse(self):
        """
        :return: A list of:

         * out_dict: a dictionary with parsed data.

         * trajectory_data: dictionary of trajectory data.

         * structure data: dictionary of cell, positions and symbols.

         * exit_code: exit code indicating potential problems of the run
           2 different keys to check in out_dict: *parser_warning* and *warnings*.

        """
        parser_version = __version__
        parser_info = {}
        parser_info["parser_warnings"] = []
        parser_info["parser_info"] = "AiiDA CASTEP basic Parser v{}".format(
            parser_version)
        parser_info["warnings"] = []

        exit_code = 'UNKNOWN_ERROR'
        finished_run = False

        # Use Total time as a mark for completed run
        for line in self.dot_castep_lines[-20:]:
            # Check only the last 20 lines
            # Otherwise unfinished restarts may be seen as finished
            if "Total time" in line:
                finished_run = True
                break

        self.parse_dot_castep()

        # Warn if the run is not finished
        if self.md_geom_info is not None:
            glines = self.md_geom_info[1]
            geom_data = self.parse_geom(glines)
            # For geom file the second energy is the enthalpy while
            # for MD it is the approx hamiltonian (etotal + ek)
            if "geom" in self.md_geom_info[0]:
                geom_data["geom_enthalpy"] = geom_data["hamilt_energy"]

        # Parse the bands file
        if self.bands_lines is not None:
            bands_res = self.parse_dot_bands()
        else:
            bands_res = None

        # Return the most 'specific' error. For example one calculation
        # may not terminate correctly due unconverged SCF.
        # In this case, we set the exit code to SCF unconverged.
        if finished_run:
            exit_code = 'CALC_FINISHED'
        else:
            exit_code = 'ERROR_NO_END_OF_CALCULATION'
            for code in EXIT_CODES_SPEC:
                if code in self.warnings:
                    exit_code = code
                    break

        # Pick out trajectory data and structure data
        if self.dot_castep_traj:
            traj_data = dict(self.dot_castep_traj)
            self.traj_data = traj_data
            if self.md_geom_info is not None:
                # Use the geom file's trajectory instead
                # some the files are overwritten
                traj_data.update(geom_data)

            try:
                last_cell = traj_data["cells"][-1]
                last_positions = traj_data["positions"][-1]
                symbols = traj_data["symbols"]

            except (KeyError, IndexError):
                # Cannot find the last geometry data
                structure_data = {}
            else:
                structure_data = dict(cell=last_cell,
                                      positions=last_positions,
                                      symbols=symbols)
        else:
            traj_data = {}
            structure_data = {}

        # A dictionary for the results to be returned
        results_dict = dict(self.dot_castep_data)
        results_dict.update(parser_info)

        # Combine the warnings
        all_warnings = results_dict["warnings"] + parser_info[
            "warnings"] + self.warnings

        # Make the warnings set-like e.g we don't want to repeat messages
        # Save a bit of the storage space
        all_warnings = list(set(all_warnings))
        results_dict['warnings'] = all_warnings

        return [results_dict, traj_data, structure_data, bands_res, exit_code]

    def parse_dot_bands(self):
        """Parse the dot_bands"""
        bands_info, kpoints, bands = parse_dot_bands(self.bands_lines)
        self.bands_res = {
            'bands_info': bands_info,
            'kpoints': kpoints,
            'bands': bands
        }
        return self.bands_res

    def parse_geom(self, glines=None):
        """Parse the geom lines"""
        if glines is None:
            glines = self.md_geom_info[1]
        self.geom_data = parse_geom_text_output(self.md_geom_info[1], {})
        return self.geom_data

    def parse_dot_castep(self):
        """Parse the dot-castep file"""
        parsed_data, trajectory_data, critical_message = parse_castep_text_output(
            self.dot_castep_lines, {})
        self.dot_castep_data = parsed_data
        self.dot_castep_traj = trajectory_data
        self.dot_castep_critical_message = critical_message
        self.warnings = parsed_data['warnings']
        return parsed_data, trajectory_data, critical_message


# Re for getting unit
unit_re = re.compile(r"^ output\s+(\w+) unit *: *([^\s]+)")
time_re = re.compile(r"^(\w+) time += +([0-9.]+) s$")
parallel_re = re.compile(
    r"^Overall parallel efficiency rating: [\w ]+ \(([0-9]+)%\)")
version_re = re.compile(r"CASTEP version ([0-9.]+)")


def parse_castep_text_output(out_lines, input_dict):
    """
    Parse ouput of .castep
    :param out_lines: a list of lines from readlines function

    :param input_dict: Control some variables. Currently support
      'n_warning_lines'- number of the lines to include for a general warning.

    :return: A list of parsed_data, trajectory_data and critical_messages:

     * parsed_data: dictionary with key values, referring to the last
       occurring quantities

     * trajectory_data: key, values of the intermediate scf steps,
       such as during geometryoptimization

     * critical_messages: a list with critical messages.

    If any is found in parsed_data["warnings"] the calculation should be considered as failed.
    """

    pseudo_files = {}
    parsed_data = {}
    parsed_data["warnings"] = []

    if not input_dict:
        input_dict = {}

    # Some parameters can be controlled
    n_warning_lines = input_dict.get("n_warning_lines", 10)

    # Initialise storage space for trajectory
    # Not all is needed. But we parse as much as we can here
    trajectory_data = defaultdict(list)

    # For the warnings I use dictionary in the format of
    # format {<keywords in line>:<message to pass>}, the message to pass
    # is save in the dictionary and used to set the exit code
    critical_warnings = {
        "SCF cycles performed but system has not reached the groundstate":
        SCF_FAILURE_ERROR,
        "STOP keyword detected in parameter file. Stop execution.":
        STOP_REQUESTED_ERROR,
        "Insufficient time for another iteration": INSUFFICENT_TIME_ERROR,
    }

    # Warnings that won't result in a calculation in FAILED state
    minor_warnings = {
        "Warning": None,
        "WARNING": None,
        "Geometry optimization failed to converge": GEOM_FAILURE_MESSAGE,
    }

    # A dictionary witch keys we should check at each line
    all_warnings = dict(critical_warnings)
    all_warnings.update(minor_warnings)

    # Split header and body
    body_start = None
    version = None
    for i, line in enumerate(out_lines):

        # Find the castep version
        if version is None:
            vmatch = version_re.search(line)
            if vmatch:
                version = vmatch.group(1)
                parsed_data["castep_version"] = version

        # Finding the units we used
        unit_match = unit_re.match(line)
        if unit_match:
            uname = unit_match.group(1)
            uvalue = unit_match.group(2)
            parsed_data["unit_" + uname] = uvalue

        if "Files used for pseudopotentials" in line:
            for line_ in out_lines[i + 1:]:
                if "---" in line_:
                    break
                try:
                    specie, pp_file = line_.strip().split()
                except ValueError:
                    break
                else:
                    pseudo_files.update({specie: pp_file})
        if "Total number of ions" in line:
            parsed_data["num_ions"] = int(line.strip().split("=")[1].strip())
            continue

        if "Point group of crystal" in line:
            parsed_data["point_group"] = line.strip().split("=")[1].strip()
            continue

        if "Space group of crystal" in line:
            parsed_data["space_group"] = line.strip().split("=")[1].strip()
            continue

        if "Cell constraints" in line:
            parsed_data["cell_constraints"] = line.strip().split(
                ":")[1].strip()
            continue

        if "Number of kpoints used" in line:
            parsed_data["n_kpoints"] = line.strip().split("=")[1].strip()
            continue

        if "MEMORY AND SCRATCH DISK ESTIMATES" in line:
            body_start = i
            break

    parsed_data.update(pseudo_pots=pseudo_files)

    # Parse repeating information
    skip = 0
    body_lines = out_lines[body_start:]

    iter_parser = get_iter_parser()
    for count, line in enumerate(body_lines):

        # Allow skipping certain number of lines
        if skip > 0:
            skip -= 1
            continue

        res_tmp = iter_parser.parse(line)
        if res_tmp:
            name, value = res_tmp[:2]
            trajectory_data[name].append(value)
            continue

        if "Calculation parallelised over" in line:
            num_cores = int(line.strip().split()[-2])
            parsed_data["parallel_procs"] = num_cores
            continue

        if "Stress Tensor" in line:
            i, stress, pressure = parse_stress_box(body_lines[count:count +
                                                              20])
            assert len(stress) == 3
            if "Symmetrised" in line:
                prefix = "symm_"
            else:
                prefix = ""
            trajectory_data[prefix + "pressure"].append(pressure)
            trajectory_data[prefix + "stress"].append(stress)
            skip = i

        if "Forces *******" in line:
            num_lines = parsed_data["num_ions"] + 10
            box = body_lines[count:(count + num_lines)]
            i, forces = parse_force_box(box)

            if "Constrained" in line:
                force_name = "cons_forces"
            else:
                force_name = "forces"
            if not forces:
                LOGGER.error("Cannot parse force lines {}".format(box))
            trajectory_data[force_name].append(forces)
            skip = i
            continue

        for warn_key in all_warnings:
            if warn_key in line:
                message = all_warnings[warn_key]
                if message is None:
                    message = body_lines[count:count + n_warning_lines]
                    message = "\n".join(message)
                parsed_data["warnings"].append(message)

    # Parse the end a few lines
    for line in body_lines[-50:]:
        time_line = time_re.match(line)
        # Save information about time usage
        if time_line:
            time_name = time_line.group(1).lower() + "_time"
            parsed_data[time_name] = float(time_line.group(2))
            continue

        para_line = parallel_re.match(line)
        if para_line:
            parsed_data["parallel_efficiency"] = int(para_line.group(1))
            continue

    #### END OF LINE BY LINE PARSING ITERATION ####

    # remove unrelated units
    units_to_delete = []
    for key in parsed_data:
        if "unit_" in key:
            unit_for = key.split("_", 1)[1]
            delete = True
            # Check the thing this unit refers do exists
            for i in trajectory_data:
                if i == key:
                    continue
                if unit_for in i:
                    delete = False

            for i in parsed_data:
                if i == key:
                    continue
                if unit_for in i:
                    delete = False

            if delete is True:
                units_to_delete.append(key)

    for key in units_to_delete:
        parsed_data.pop(key)

    # set geom convergence state
    if GEOM_FAILURE_MESSAGE in parsed_data["warnings"]:
        parsed_data["geom_unconverged"] = True
    else:
        parsed_data["geom_unconverged"] = None

    return parsed_data, trajectory_data, list(critical_warnings.values())


class LineParser:
    """
    Parser for a line
    """
    def __init__(self, conditions):
        """initialize the Parser by passing the conditions"""
        self._cond = conditions

    def parse(self, line):
        """
        Return parsing results

        :returns: Result of the first matched Matcher object or None if no match is found
        """
        out = None
        for c in self._cond:
            out, _ = c.match_pattern(line)
            if out:
                break
        return out


class Matcher:
    """
    Class of the condition to match the line
    """
    def __init__(self, regex, name, convfunc=None):
        """
        Initialize a Matcher object.

        :param string regex: Pattern to be matched
        :param string name: Name of the results
        """
        self.regex = re.compile(regex)
        self.convfunc = convfunc
        self.name = name

    def match_pattern(self, line):
        """
        Match pattern

        :returns: (out, match) Out is a dicationary of {self.name: <matched_value>}.
          and match is a re.MatchObject or None
        """

        match = self.regex.match(line)
        if match:
            value = match.group(1)
            if self.convfunc:
                value = self.convfunc(value)
            out = (self.name, value)
        else:
            out = None
        return out, match


class UnitMatcher(Matcher):
    """
    The pattern of a UnitMatcher should have two groups with second group
    being the unit. The first group will be converted to float
    """
    def match_pattern(self, line):
        """
        Match the pattern
        """
        out, match = super(UnitMatcher, self).match_pattern(line)
        if out:
            unit = match.group(2)
            conv = self.convfunc if self.convfunc else float
            out = (out[0], conv(out[1]), unit)
        return out, match


def get_iter_parser():
    """
    Generate a LineParser object to parse repeating outputs
    """
    tail1 = r' *= *([0-9.+-eE]+) +(\w+)'
    mfree = UnitMatcher(r'^Final free energy \(E-TS\)' + tail1, "free_energy")
    mtotal = UnitMatcher(r'^Final energy, E' + tail1, "total_energy")
    mtotal2 = UnitMatcher(r'^Final energy' + tail1, "total_energy")
    mzeroK = UnitMatcher(r'^NB est. 0K energy \(E-0.5TS\)' + tail1,
                         "zero_K_energy")
    spin = UnitMatcher(r'^Integrated Spin Density' + tail1, "spin_density")
    absspin = UnitMatcher(r'^Integrated \|Spin Density\|' + tail1,
                          "abs_spin_density")
    enthalpy = UnitMatcher(
        r'^ *\w+: finished iteration +\d+ +with enthalpy' + tail1, "enthalpy")
    parser = LineParser(
        [mfree, mtotal, mtotal2, mzeroK, spin, absspin, enthalpy])
    return parser


def parse_geom_text_output(out_lines, input_dict):
    """
    Parse output of .geom file

    :param out_lines: a list of lines from the readline function.
    :param dict input_dict: not in use at the moment.

    :return: key, value of the trajectories of cell, atoms, force etc
    """
    _ = input_dict

    txt = out_lines
    Hartree = units['Eh']  # eV
    Bohr = units['a0']  # A
    kB = units['kB']  # eV/K
    hBar = units['hbar']  # in eV
    eV = units["e"]  # in J

    cell_list = []
    species_list = []
    geom_list = []
    forces_list = []
    energy_list = []
    hamilt_list = []
    kinetic_list = []
    pressure_list = []
    temperature_list = []
    velocity_list = []
    time_list = []

    # For a specific image
    current_pos = []
    current_species = []
    current_forces = []
    current_velocity = []
    current_cell = []
    in_header = False
    for line in txt:
        if "begin header" in line.lower():
            in_header = True
            continue
        if "end header" in line.lower():
            in_header = False
            continue
        if in_header:
            continue  # Skip header lines

        sline = line.split()
        if len(sline) == 1:
            try:
                time_list.append(float(sline[0]))
            except ValueError:
                continue
        elif '<-- E' in line:
            energy_list.append(float(sline[0]))  # Total energy
            hamilt_list.append(float(sline[1]))  # Hamitonian (MD)
            # Kinetic energy is not blank in GEOM OPT runs
            if len(sline) == 5:
                kinetic_list.append(float(sline[2]))  # Kinetic (MD)
            continue
        elif '<-- h' in line:
            current_cell.append(list(map(float, sline[:3])))
            continue
        elif '<-- R' in line:
            current_pos.append(list(map(float, sline[2:5])))
            current_species.append(sline[0])
        elif '<-- F' in line:
            current_forces.append(list(map(float, sline[2:5])))
        elif '<-- V' in line:
            current_velocity.append(list(map(float, sline[2:5])))
        elif '<-- T' in line:
            temperature_list.append(float(sline[0]))
        elif '<-- P' in line:
            pressure_list.append(float(sline[0]))
        elif not line.strip() and current_cell:
            cell_list.append(current_cell)
            species_list.append(current_species)
            geom_list.append(current_pos)
            forces_list.append(current_forces)
            current_cell = []
            current_species = []
            current_pos = []
            current_forces = []
            if current_velocity:
                velocity_list.append(current_velocity)
                current_velocity = []

    if len(species_list) == 0:
        raise CASTEPOutputParsingError("No data found in geom file")

    out = dict(
        cells=np.array(cell_list) * Bohr,
        positions=np.array(geom_list) * Bohr,
        forces=np.array(forces_list) * Hartree / Bohr,
        geom_total_energy=np.array(energy_list) * Hartree,
        symbols=species_list[0],
    )

    # optional lists
    unit_V = Hartree * Bohr / hBar
    unit_K = Hartree / kB  # temperature in K
    unit_P = Hartree / (Bohr * 1e-10)**3 * eV
    unit_s = hBar / Hartree
    opt = {
        "velocities": (velocity_list, unit_V),
        "temperatures": (temperature_list, unit_K),
        "pressures": (pressure_list, unit_P),
        "hamilt_energy": (hamilt_list, Hartree),
        "times": (time_list, unit_s),
        "kinetic_energy": (kinetic_list, Hartree)
    }
    for key, value in opt.items():
        if value[0]:
            out.update({key: np.array(value[0]) * value[1]})
    return out


force_match = re.compile(
    r"^ +\* +(\w+) +([0-9]+) +([0-9.\-+]+) +([0-9.\-+]+) +([0-9.\-+]+) +\*")
stress_match = re.compile(
    r"^ +\* +(\w+) +([0-9.\-+]+) +([0-9.\-+]+) +([0-9.\-+]+) +\*")


def parse_force_box(lines):
    """
    Parse a box of the forces
    :param lines: a list of upcoming lines

    :return: A list of number of lines read and the forces
    """

    forces = []
    i = 0
    for i, line in enumerate(lines):

        if "Forces" in line:
            continue

        if "***********************" in line:
            break
        # remove the cons'd tags
        line = line.replace('(cons\'d)', '        ')

        match = force_match.match(line)
        if match:
            forces.append([float(match.group(i)) for i in range(3, 6)])

    return i, forces


def parse_stress_box(lines):
    """
    Parse a box of the stress
    :param lines: a list of upcoming lines

    :return: a list of  [number of lines read, stress_tensor, pressure]
    """

    stress = []
    pressure = None
    i = 0
    for i, line in enumerate(lines):
        if "Stress" in line:
            continue

        if "Cartisian components" in line:
            unit = line.strip().split()[-2]
            unit = unit[1:-2]
            continue

        if "**********************" in line:
            break

        match = stress_match.match(line)
        if match:
            stress.append([float(match.group(i)) for i in range(2, 5)])
        elif "Pressure" in line:
            lsplit = line.strip().split()
            pressure = float(lsplit[-2])

    return i, stress, pressure


def parse_dot_bands(bands_lines):
    """
    Parse an CASTEP bands file
    Extract Kpoints and each bands for each kpoints.
    This is a generic parsing function. Return python builtin types.

    The problems is the order of the kpoints written is parallelisation
    dependent and may not be the same as that specified in *kpoints_list*.
    However, CASTEP does write the index of the kpoints.
    We reorder the kpoints and the bands to match the original order in
    the input file.


    :param bands_lines: A list of lines to be parsed parse
    :return: A list of bands_info, kpoints and bands:

     * bands_info: A dictionary for information of bands.
     * kpoints: A list of kpoints. In the format [kpoint index,
         coordinats x 3 kpoint weight]
     * bands: A list of bands. Each band has a list of actual
       eigenvalues for each spin components. E.g nkpoints, nspins, neigns.

    Note that the atomic units are used in the bands file
    """
    i_finish = None
    cell = []
    bands_info = {}
    i = 0
    for i, line in enumerate(bands_lines):
        if not line.strip():
            continue
        if "Number of k-points" in line:
            nkps = line.strip().split()[-1]
            bands_info['nkpts'] = int(float(nkps))
            continue
        if "Number of spin components" in line:
            nspin = line.strip().split()[-1]
            bands_info['nspins'] = int(float(nspin))
            continue
        if "Number of electrons" in line:
            nelec = line.strip().split()[-1]
            bands_info['nelecs'] = float(nelec)
            continue
        if "Number of eigenvalues" in line:
            neigns = line.strip().split()[-1]
            bands_info['neigns'] = int(float(neigns))
            continue
        if "Fermi energ" in line:  # NOTE possible different format with spin polarisation?
            efermi = line.strip().split()[-1]
            bands_info['efermi'] = float(efermi)
            continue
        if "Unit cell" in line:
            i_finish = i + 3
            continue

        # Added the cell
        if i_finish:
            cell.append([float(n) for n in line.strip().split()])
        if i == i_finish:
            break
    bands_info['cell'] = cell

    # Now parse the body
    kpoints = []
    bands = []
    this_band = []
    this_spin = []
    for line in bands_lines[i + 1:]:
        if "K-point" in line:
            # We are not at the first kpoints
            if kpoints:
                # Save the result from previous block
                this_band.append(this_spin)
                bands.append(this_band)
                this_spin = []
                this_band = []
            kline = line.strip().split()
            kpoints.append([float(n) for n in kline[1:]])
            continue
        if "Spin component" in line:
            # Check if we are at the second spin
            if this_spin:
                this_band.append(this_spin)
                this_spin = []
            continue

        ls = line.strip()
        if not ls:
            continue
        this_spin.append(float(ls))

    # Save the last set of results
    this_band.append(this_spin)
    bands.append(this_band)

    # Do some sanity checks
    assert int(nkps) == len(kpoints), "Missing kpoints"
    assert len(bands) == len(kpoints), "Missing bands for certain kpoints"

    for n, b in enumerate(bands):
        assert len(b) == int(nspin), "Missing spins for kpoint {}".format(n +
                                                                          1)
        for i, s in enumerate(b):
            assert len(s) == int(neigns), ("Missing eigenvalues "
                                           "for kpoint {} spin {}".format(
                                               n + 1, i + 1))

    return bands_info, kpoints, bands



#!/usr/bin/env python
"""
Mock running CASTEP
"""

from __future__ import absolute_import
from __future__ import print_function
import hashlib
import json
import tempfile
import click
import shutil

import castepinput
from aiida_castep import tests
from pathlib import Path

_TEST_BASE_FOLDER = Path(tests.__file__).parent
_TEST_DATA_FOLDER = _TEST_BASE_FOLDER / 'data'


def get_hash(dict_obj):
    """
    Return hash of a dict of strings
    """

    rec = []
    for k, v in dict_obj.items():
        if isinstance(v, str):
            rec.append(k + ":" + v)
        elif isinstance(v, list):
            rec.append(k + "".join(v))
    # Update, use sorted so the original order does not matter
    base = [record.encode().lower() for record in sorted(rec)]
    # Compute the hash
    md5 = hashlib.md5()
    [md5.update(b) for b in base]
    return md5.hexdigest(), base


def test_hash():
    """Test hashing"""
    exmp = {'a': 'b', 'b': 'a'}
    get_hash(exmp)

    exmp = {'a': 'b', 'b': ['a', 'c']}
    h1, b1 = get_hash(exmp)

    exmp = {'b': ['a', 'c'], 'a': 'b'}
    h2, b2 = get_hash(exmp)
    assert b1 == b2
    assert h1 == h2


class MockOutput(object):
    def __init__(self, base_dir=None):
        """
        Initialize the object
        """
        if base_dir is None:
            self.base_dir = _TEST_DATA_FOLDER
        else:
            self.base_dir = Path(base_dir).resolve()

    def calculate_hash(self, path):

        path = str(path)
        param = castepinput.ParamInput.from_file(path + '.param', plain=True)
        cell = castepinput.CellInput.from_file(path + '.cell', plain=True)
        all_inp = dict(param)
        all_inp.update(cell)
        all_inp.pop('comment', None)
        all_inp.pop('COMMENT', None)
        return get_hash(all_inp)[0]

    @property
    def _reg_file(self):
        """
        Path to the registry file
        """
        return self.base_dir / 'registry.json'

    @property
    def registry(self):
        """
        Registry, a dictionary loaded from the json
        """
        if not self._reg_file.is_file():
            return {}

        with open(str(self._reg_file), 'r') as fh:
            reg = json.load(fh)
        return reg

    def register_node(self, calcjob, tag=None):
        """
        Register the result of a CalcJob node
        :param calcjob: The CalcJob node to be used
        :param tag: Tag for the results, used as the folder name
        """

        # Copy the 'objects' to an temporary directory
        # Include both the retrieved and the inputs
        tmpwork = tempfile.mkdtemp()
        retrieved = calcjob.outputs.retrieved
        for node in [retrieved, calcjob]:
            for nm in node.list_object_names():
                if nm.startswith('.') or \
                   nm.startswith('_'):
                    continue
                with node.open(nm) as fin:
                    fpath = Path(tmpwork) / nm
                    fpath.write_text(fin.read())

        seedpath = Path(tmpwork) / calcjob.get_option('seedname')
        # Register the resutls
        self.register(seedpath, tag)

        shutil.rmtree(tmpwork)

    def register(self, seedpath, tag=None):
        """
        Register completed calculation. Such calculation must be in the directroy
        tree of the mock_castep.py
        """
        seedpath = Path(seedpath).resolve()
        hash_ = self.calculate_hash(seedpath)

        if self._reg_file.is_file():
            try:
                with open(str(self._reg_file)) as fh:
                    reg = json.load(fh)
            except RuntimeError:
                reg = {}
        else:
            reg = {}

        # Relative seedpath to the folder
        try:
            rel_path = seedpath.relative_to(self.base_dir).parent
        except ValueError:
            # It is not placed in the sub folder
            # Copy manually
            import shutil
            if tag is None:
                # Use the first 8 digits as the name
                tag = hash_[:8]

            shutil.copytree(str(seedpath.parent), str(self.base_dir / tag))
            rel_path = tag
        reg[hash_] = str(rel_path)

        # Save the json settings
        with open(str(self._reg_file), 'w') as fh:
            json.dump(reg, fh)

    def copy_results(self, rel_path):
        """
        Copy existing calculation to the folder
        """
        print('Selected path:', rel_path)
        import shutil
        res_files = (self.base_dir / rel_path).glob('*')
        cwd = Path.cwd()
        for r in res_files:
            shutil.copy(str(r), str(cwd))
        return

    def run(self, seedname):
        """
        Run the 'Calculation', if the seed is known we just copy
        the results
        Can be overiden by the 'MOCK_CALC' environmental varaible
        , which define the folder of results to be copied
        """
        import os
        overide = os.environ.get('MOCK_CALC')
        if overide:
            self.copy_results(overide)
            print('Overiden by MOCK_CALC')
            print('Returning results from {}'.format(Path(overide).resolve()))
            return

        hash_ = self.calculate_hash(seedname)
        reg = self.registry

        known_result = reg.get(hash_, None)
        if known_result:
            self.copy_results(known_result)
            print('Returning results from {}'.format(self.base_dir /
                                                     known_result))
        else:
            raise RuntimeError('Results not registered')


@click.command('mock')
@click.option('--reg',
              default=False,
              is_flag=True,
              help='Register the calculation')
@click.option('--tag',
              help='Tag for the folder when registering results',
              default=None)
@click.argument('seed')
def main(seed, reg, tag):
    runner = MockOutput()
    if reg:
        runner.register(seed, tag=tag)
    else:
        runner.run(seed)


if __name__ == "__main__":
    main()



"""
Utility functions
"""
import copy

import numpy as np
from aiida.common import OutputParsingError
from aiida.plugins import DataFactory


class CASTEPOutputParsingError(OutputParsingError):
    pass


def structure_from_input(cell, positions, symbols):
    """
    Receives the dictionary cell parsed from CASTEP
    Convert it into an AiiDA structure object
    """

    SructureData = DataFactory("structure")  # pylint: disable=invalid-name

    out_structure = SructureData(cell=cell)

    for symbol, position in zip(symbols, positions):

        out_structure.append_atom(symbols=symbol, position=position)

    return out_structure


def add_last_if_exists(dict_of_sequence, key, dict_to_be_added):
    """
    Added the last term of a sequence to a dictionary.
    This is used for collecting final values in a dictionary of 'trajectory'
    """

    try:
        last = dict_of_sequence[key][-1]
    except (KeyError, IndexError):
        return
    # Check if last exist - in case of DefaultDict being passed
    if last:
        dict_to_be_added[key] = last


def desort_structure(structure, original_structure):
    """
    Recover the order of structure. CASTEP will sort the input structure
    according to the atomic numbers
    """
    new_structure = copy.deepcopy(structure)

    rsort = get_desort_args(original_structure)
    new_sites = np.array(structure.sites)[rsort]

    # Map back to the original order
    new_structure.clear_sites()
    for site in new_sites:
        new_structure.append_site(site)

    # Check for sure
    assert [site.kind_name for site in original_structure.sites
            ] == [site.kind_name for site in new_structure.sites]

    return new_structure


def get_desort_args(original_structure):
    """
    Return an index array for desorting the structure

    :return: An array used to recovery the original order
    """
    numbers = original_structure.get_ase().numbers
    isort = np.argsort(numbers, kind='mergesort')
    rsort = [-1] * len(numbers)
    for idx, value in enumerate(isort):
        rsort[value] = idx
    assert -1 not in rsort

    return rsort



"""
Classes for .param and .cell files
"""
from __future__ import absolute_import
from collections import OrderedDict


class CastepInputFile(OrderedDict):
    """
    Class for storing key - values pairs of CASTEP inputs
    This class can be used for .param, .cell and also other CASTEP style
    inputs such as OptaDos's odi file

    ``self.get_file_lines`` is used for getting a list of strings as
    lines to be written to the file

    ``self.get_string`` is used for getting the content to be passed
    to ``write`` function of a file-like object

    sepecial properties:
    * ``header`` a list of lines to be put into the header
    * ``units`` a dictionary of the units
    """
    def __init__(self, *args, **kwargs):
        super(CastepInputFile, self).__init__(*args, **kwargs)
        self.header = []
        self.units = {}

    def get_file_lines(self):
        """
        Return a list of strings to be write out to the files
        """
        lines = []
        for head in self.header:
            if not head.startswith("#"):
                lines.append("# " + head)
            else:
                lines.append(head)

        for key, value in self.items():
            if isinstance(value, (tuple, list)):
                lines.append("%BLOCK {}".format(key))
                if key in self.units:
                    lines.append("{}".format(self.units[key]))
                for tmp in value:
                    lines.append(tmp)
                lines.append("%ENDBLOCK {}".format(key))
            else:
                line = "{:<20}: {}".format(key, value)
                if key in self.units:
                    line = line + " " + self.units[key]
                lines.append(line)

        return lines

    def get_string(self):
        return "\n".join(self.get_file_lines())


class ParamFile(CastepInputFile):
    pass


class CellFile(CastepInputFile):
    pass



"""
Module for generation HELP information using CASTEP
excutable.
"""
from __future__ import print_function
from __future__ import absolute_import
import subprocess as sbp
import re
import sys


def dot_proc(iterable):
    """Provide a primitive progress bar"""
    print("Processing keywords")
    m = len(iterable) - 1
    print("=" * (m // 10))
    for n, i in enumerate(iterable):
        if n % 10 == 0:
            print(".", end="")
            sys.stdout.flush()
        if n == m:
            print("\n")
        yield i


def progress(func, *args, **kwargs):
    try:
        from tqdm import tqdm
    except ImportError:
        return dot_proc(func)
    else:
        return tqdm(func, *args, **kwargs)


def get_castep_commands(castep_command="castep.serial", key="all"):

    outlines = sbp.check_output([castep_command, "-h", key],
                                universal_newlines=True)
    lines = outlines.split("\n")

    cell = {}
    param = {}
    for i, line in enumerate(lines):
        # Skip empty lines
        if not line.strip():
            continue
        if "Help information on CELL keywords" in line:
            continue
        if "Help information on PARAMETERS keywords" in line:
            next_section = i
            break

        key, helpinfo = line.strip().split(None, 1)
        cell[key] = helpinfo

    for line in lines[next_section:]:

        if not line.strip():
            continue

        key, helpinfo = line.strip().split(None, 1)
        param[key] = helpinfo

    return cell, param


allowed_value_re = re.compile("Allow values: ([^.\n]+)[.\n]")
default_value_re = re.compile("Default value: ([^.\n]+)[.\n]")
type_re = re.compile("Type: (\w+)")
level_re = re.compile("Level: (\w+)")


def parse_help_string(key, excutable="castep.serial"):
    """Capture help string, determine if it is for PARAM or CELL"""

    out = sbp.check_output([excutable, "-h", key], universal_newlines=True)
    lines = out.split("\n")
    value_type = None
    key_level = None

    for i, line in enumerate(lines):
        if "Help information on PARAMETERS keywords" in line:
            param_start = i

        match = type_re.search(line)
        if match and not value_type:
            value_type = match.group(1).lower()

        match = level_re.search(line)
        if match and not key_level:
            key_level = match.group(1).lower()

    cell_lines = lines[2:param_start]
    param_lines = lines[param_start + 2:]

    if len(cell_lines) > len(param_lines):
        help_lines = cell_lines
        key_type = "CELL"
    else:
        help_lines = param_lines
        key_type = "PARAM"

    return help_lines, key_type, key_level, value_type



"""
Tests for data module
"""
from __future__ import absolute_import
from aiida_castep.data.utils import split_otfg_entry, get_usp_element

Ti_otfg = "Ti 3|1.8|9|10|11|30U:40:31:32(qc=5.5)"
Sr_otfg = "Sr 3|2.0|5|6|7|40U:50:41:42"
O_otfg = "O 2|1.1|15|18|20|20:21(qc=7)"
raw_otfg = "2|1.1|15|18|20|20:21(qc=7)"
otfglibs = ["QC5", "NCP17", "C9"]


def test_otfg_split():
    element, setting = split_otfg_entry(Ti_otfg)
    assert element == "Ti"
    assert setting == "3|1.8|9|10|11|30U:40:31:32(qc=5.5)"

    O, O_entry = split_otfg_entry(O_otfg)
    elem, O_entry2 = split_otfg_entry(raw_otfg)
    assert elem is None
    assert O_entry2 == O_entry

    for otfglib in otfglibs:
        elem, entry = split_otfg_entry(otfglib)
        assert entry == otfglib
        assert elem == "LIBRARY"


def test_usp_element():
    """Test the element is currectly parsed from a file path"""

    fpath = "/tmp/Sr_00.usp"
    elem = get_usp_element(fpath)
    assert elem == "Sr"

    fpath = "/tmp/sr_00.usp"
    elem = get_usp_element(fpath)
    assert elem == "Sr"

    elem = get_usp_element('/tmp/Sr_00.upf')
    assert elem is None



from __future__ import absolute_import
import unittest
import pytest
from aiida_castep.parsers.raw_parser import (parse_castep_text_output,
                                             parse_geom_text_output,
                                             parse_dot_bands, RawParser)

from pathlib import Path
import os


class TestParsers(unittest.TestCase):
    @property
    def data_abs_str(self):
        test_moudule = Path(__file__).parent
        data_folder = test_moudule / 'data'
        return str(data_folder)

    @property
    def data_abs_path(self):
        test_moudule = Path(__file__).parent
        data_folder = test_moudule / 'data'
        return data_folder

    def get_lines(self, path):
        """Get a list of lines from a path"""
        with open(str(path)) as fh:
            lines = fh.readlines()
        return lines

    def setUp(self):
        with open(self.data_abs_str + "/H2-geom/aiida.geom") as fh:
            self.geom_lines = fh.readlines()

        with open(self.data_abs_str + "/H2-geom/aiida.castep") as cs:
            self.castep_lines = cs.readlines()

        pass

    def test_parse_geom(self):
        res = parse_geom_text_output(self.geom_lines, None)
        self.assertEqual(res["symbols"], ["H", "H"])
        self.assertEqual(res["geom_total_energy"].shape[0], 5)
        self.assertEqual(res["positions"].shape[0], 5)
        self.assertEqual(res["forces"].shape[0], 5)

    def test_parse_castep(self):
        parsed_data, trajectory_data, warnings = parse_castep_text_output(
            self.castep_lines, None)
        self.assertTrue(parsed_data["cell_constraints"])
        self.assertFalse(parsed_data["warnings"])
        self.assertTrue(parsed_data["n_kpoints"])
        self.assertTrue(parsed_data["space_group"])
        self.assertTrue(parsed_data["pseudo_pots"])
        self.assertTrue(parsed_data["point_group"])
        self.assertTrue(trajectory_data["enthalpy"])
        self.assertTrue(trajectory_data["total_energy"])
        self.assertLess(
            trajectory_data["total_energy"][0] -
            trajectory_data["enthalpy"][0], 1e-5)
        self.assertFalse(parsed_data["warnings"])
        self.assertEqual(parsed_data["total_time"], 14.53)
        self.assertEqual(parsed_data["initialisation_time"], 1.02)
        self.assertEqual(parsed_data["parallel_efficiency"], 90)
        self.assertEqual(parsed_data["castep_version"], "17.2")

    def test_warnings(self):
        # Test assertion of warnings
        with_warning = self.castep_lines[:]
        # This is no longer a critical warning leading to FAILED state
        with_warning.insert(-10, "Geometry optimization failed to converge")
        parsed_data, trajectory_data, critical = parse_castep_text_output(
            with_warning, None)
        self.assertTrue(parsed_data["warnings"])
        self.assertNotIn(parsed_data["warnings"][0], critical)

        with_warning = self.castep_lines[:]
        with_warning.insert(
            -10,
            "SCF cycles performed but system has not reached the groundstate")
        parsed_data, trajectory_data, critical = parse_castep_text_output(
            with_warning, None)
        self.assertTrue(parsed_data["warnings"])
        self.assertIn(parsed_data["warnings"][0], critical)

    def test_parse_bands(self):
        """
        Test the function to parse *.bands file
        """
        res = parse_dot_bands(
            self.get_lines(self.data_abs_path / 'Si-geom-stress/aiida.bands'))
        self.assertEqual(res[0]['nspins'], 1)
        self.assertEqual(res[0]['nkpts'], len(res[1]))
        self.assertEqual(res[0]['neigns'], len(res[2][0][0]))

    def test_parser_stress(self):
        with open(self.data_abs_str + "/Si-geom-stress/aiida.castep") as cs:
            lines = cs.readlines()
        parsed_data, trajectory_data, warnings = parse_castep_text_output(
            lines, None)

        self.assertIn('symm_stress', trajectory_data)
        self.assertIn('symm_pressure', trajectory_data)
        self.assertTrue(trajectory_data['symm_pressure'])
        self.assertTrue(len(trajectory_data['symm_pressure']) > 10)

    def test_parser_class(self):
        """Test the classfor RawParser"""
        bands = self.get_lines(self.data_abs_path /
                               'Si-geom-stress/aiida.bands')
        parser = RawParser(self.castep_lines, {},
                           ['aiida.geom', self.geom_lines], bands)
        res = parser.parse()



#!/usr/bin/env python

from __future__ import absolute_import
import subprocess
import os
from os.path import abspath, dirname, join
from setuptools import setup, find_packages
import json


def git_version():
    def _minimal_ext_cmd(cmd):
        # construct minimal environment
        env = {}
        for k in ['SYSTEMROOT', 'PATH', 'HOME']:
            v = os.environ.get(k)
            if v is not None:
                env[k] = v
        # LANGUAGE is used on win32
        env['LANGUAGE'] = 'C'
        env['LANG'] = 'C'
        env['LC_ALL'] = 'C'
        out = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                               env=env).communicate()[0]
        return out

    try:
        out = _minimal_ext_cmd(['git', 'rev-parse', '--short', 'HEAD'])
        GIT_REVISION = out.strip().decode('ascii')
    except OSError:
        GIT_REVISION = "Unknown"

    return GIT_REVISION


if __name__ == '__main__':

    # Provide static information in setup.json
    # such that it can be discovered automatically
    ROOT = abspath(dirname(__file__))
    with open(join(ROOT, 'setup.json'), 'r') as info:
        kwargs = json.load(info)

    # Check if in a CI environment
    is_tagged = False
    if os.environ.get('CI_COMMIT_TAG'):
        ci_version = os.environ['CI_COMMIT_TAG']
        is_tagged = True
    elif os.environ.get('CI_JOB_ID'):
        ci_version = os.environ['CI_JOB_ID']
    else:
        # Note in CI
        ci_version = None

    if ci_version:
        # If this a release, check the consistency
        if is_tagged:
            assert ci_version == kwargs[
                'version'], 'Inonsistency between versions'
        else:
            kwargs['version'] = ci_version

# Included the README.md as the long description
    with open(join(ROOT, 'README.md'), 'r') as f:
        long_desc = f.read()
    setup(packages=find_packages(),
          long_description=long_desc,
          long_description_content_type='text/markdown',
          include_package_data=True,
          **kwargs)



"""
Store common stuff
"""

# Mapping of the input names
from __future__ import absolute_import
from collections import OrderedDict

INPUT_LINKNAMES = {
    'structure': 'structure',  # Input structure
    'parameters':
    'parameters',  # Input parameters e.g defines .cell and .param
    'kpoints': 'kpoints',  # Input kpoints
    'pseudo': 'pseudo',  # Input pseudopotential, namespace
    'settings': 'settings',  # Extra settings for CASTEP
    'parent_calc_folder':
    'parent_calc_folder',  # Remote folder point to the parent calculation
    'prod_structure':
    'product_structure'  # Product structure in transition state search
}

OUTPUT_LINKNAMES = {
    'structure': 'output_structure',  # The output structure
    'results': 'output_parameters',  # Basic infomration, energies, units, etc
    'trajectory':
    'output_trajectory',  # Trajectory of md or geometry optimisation
    'bands': 'output_bands',  # Bands of the structure
    'array': 'output_array'  # Array of values, for example SCF energies
}

# We define an ordered dictionary of the error codes
# The error are in the order of decending priority, the error with
# the highest priority is used as the return code
# The values are used to define the ExitCode which is a namedtuple with field
# status, message, invalidates_cache

EXIT_CODES_SPEC = OrderedDict((
    ('ERROR_SCF_NOT_CONVERGED', (101, 'SCF Cycles failed to reach convergence',
                                 False)),
    ('ERROR_STOP_REQUESTED',
     (103,
      'Stopped execuation due to detection of \'stop \' keyword in param file.',
      True)),
    ('ERROR_TIMELIMIT_REACHED',
     (107, 'Calculation self-terminated due to time limit', False)),
    # Errors with missing files
    ('ERROR_CASTEP_ERROR',
     (104, 'CASTEP generate error files. Check them for details', True)),
    ('ERROR_NO_END_OF_CALCULATION',
     (105, 'Cannot find the end of calculation',
      True)),  # Indicated by the lack of summary line
    ('ERROR_NO_OUTPUT_FILE', (106, 'No output .castep files found', True)),
    ('ERROR_NO_RETRIEVE_FOLDER', (108, 'No retrieve folder is found', True)),
    ('UNKOWN_ERROR', (200, 'UNKOWN ERROR', True)),
    ('CALC_FINISHED', (0, 'Calculation terminated gracefully, end found',
                       False)),
))

# exit code dictionary with the numerical code as the keys
EXIT_CODE_NUMS = OrderedDict(
    (v[0], (k, v[1], v[2])) for k, v in EXIT_CODES_SPEC.items())



"""
Parsers for CASTEP
"""
from __future__ import absolute_import
from copy import deepcopy
import six
import numpy as np

from aiida.orm import TrajectoryData, ArrayData, Dict, BandsData

from aiida.parsers.parser import Parser
from aiida.common import exceptions

from aiida_castep.parsers.raw_parser import units, RawParser
from aiida_castep.parsers.utils import (structure_from_input,
                                        add_last_if_exists, desort_structure,
                                        get_desort_args)
from aiida_castep.common import OUTPUT_LINKNAMES as out_ln
from aiida_castep.common import EXIT_CODES_SPEC as calc_exit_code
from aiida_castep._version import CALC_PARSER_VERSION

# pylint: disable=invalid-name,too-many-locals,too-many-statements,too-many-branches
__version__ = CALC_PARSER_VERSION

ERR_FILE_WARNING_MSG = ".err files found in workdir"


class CastepParser(Parser):
    """
    This is the class for Parsing results from a CASTEP calculation
    Supported calculations types:
    singlepoint
    geom
    """

    _setting_key = 'parser_options'

    def parse(self, **kwargs):
        """
        Receives a dictionary of retrieved nodes.retrieved.
        Top level logic of operation
        """

        try:
            output_folder = self.retrieved
        except exceptions.NotExistent:
            return self.exit_codes.ERROR_NO_RETRIEVED_FOLDER

        warnings = []
        exit_code_1 = None

        # NOTE parser options not used for not
        parser_opts = {}

        # NOT READILY IN USE
        input_dict = {}

        # check what is inside the folder
        filenames = [f.name for f in output_folder.list_objects()]

        # Get calculation options
        options = self.node.get_options()
        seedname = options['seedname']

        # at least the stdout should exist
        if options['output_filename'] not in filenames:
            self.logger.error("Standard output not found")
            return self.exit_codes.ERROR_NO_OUTPUT_FILE

        # The calculation is failed if there is any err file.
        err_filenames = [fname for fname in filenames if '.err' in fname]
        if err_filenames:
            exit_code_1 = 'ERROR_CASTEP_ERROR'

        # Add the content of err files
        err_contents = set()
        for fname in err_filenames:
            err_contents.add(output_folder.get_object_content(fname))

        # Trajectory files
        has_md_geom = False
        out_md_geom_name_content = None
        for suffix in ('.geom', '.md'):
            fname = seedname + suffix
            if fname in filenames:
                out_md_geom_name_content = (
                    fname, output_folder.get_object_content(fname).split('\n'))
                has_md_geom = True
                break

        # Handling bands
        fname = seedname + '.bands'
        has_bands = fname in filenames
        if has_bands:
            out_bands_content = output_folder.get_object_content(fname).split(
                '\n')
        else:
            out_bands_content = None

        out_file = options['output_filename']
        out_file_content = output_folder.get_object_content(out_file).split(
            '\n')

        ###### CALL THE RAW PASSING FUNCTION TO PARSE DATA #######

        raw_parser = RawParser(out_lines=out_file_content,
                               input_dict=input_dict,
                               md_geom_info=out_md_geom_name_content,
                               bands_lines=out_bands_content,
                               **parser_opts)
        out_dict, trajectory_data, structure_data, bands_data, exit_code_2\
            = raw_parser.parse()

        # Combine the exit codes use the more specific error
        exit_code = None
        for code in calc_exit_code:
            if code in (exit_code_2, exit_code_1):
                exit_code = code
                break

        # Append the final value of trajectory_data into out_dict
        last_value_keys = [
            "free_energy", "total_energy", "zero_K_energy", "spin_density",
            "abs_spin_density", "enthalpy"
        ]
        for key in last_value_keys:
            add_last_if_exists(trajectory_data, key, out_dict)

        # Add warnings from this level
        out_dict["warnings"].extend(warnings)

        # Add error messages
        out_dict["error_messages"] = list(err_contents)

        ######## --- PROCESSING BANDS DATA -- ########
        if has_bands:
            bands_node = bands_to_bandsdata(**bands_data)
            self.out(out_ln['bands'], bands_node)

        ######## --- PROCESSING STRUCTURE DATA --- ########
        no_optimise = False
        try:
            cell = structure_data["cell"]
            positions = structure_data["positions"]
            symbols = structure_data["symbols"]

        except KeyError:
            # Handle special case where CASTEP founds nothing to optimise,
            # hence we attached the input geometry as the output
            for warning in out_dict["warnings"]:
                if "there is nothing to optimise" in warning:
                    no_optimise = True
            if no_optimise is True:
                self.out(out_ln['structure'],
                         deepcopy(self.node.inputs.structure))
        else:
            structure_node = structure_from_input(cell=cell,
                                                  positions=positions,
                                                  symbols=symbols)
            # Use the output label as the input label
            input_structure = self.node.inputs.structure
            structure_node = desort_structure(structure_node, input_structure)
            structure_node.label = input_structure.label
            self.out(out_ln['structure'], structure_node)

        ######### --- PROCESSING TRAJECTORY DATA --- ########
        # If there is anything to save
        # It should...
        if trajectory_data:

            # Resorting indices - for recovering the original ordering of the
            # species in the input structure
            input_structure = self.node.inputs.structure
            idesort = get_desort_args(input_structure)
            # If we have .geom file, save as in a trajectory data
            if has_md_geom:
                try:
                    positions = np.asarray(
                        trajectory_data["positions"])[:, idesort]
                    cells = trajectory_data["cells"]
                    # Assume symbols do not change - symbols are the same for all frames
                    symbols = np.asarray(trajectory_data["symbols"])[idesort]
                    stepids = np.arange(len(positions))

                except KeyError:
                    out_dict["parser_warning"].append(
                        "Cannot "
                        "extract data from .geom file.")

                else:
                    traj = TrajectoryData()
                    traj.set_trajectory(stepids=np.asarray(stepids),
                                        cells=np.asarray(cells),
                                        symbols=np.asarray(symbols),
                                        positions=np.asarray(positions))
                    # Save the rest
                    for name, value in six.iteritems(trajectory_data):
                        # Skip saving empty arrays
                        if len(value) == 0:
                            continue

                        array = np.asarray(value)
                        # For forces/velocities we also need to resort the array
                        if ("force" in name) or ("velocities" in name):
                            array = array[:, idesort]
                        traj.set_array(name, np.asarray(value))
                    self.out(out_ln['trajectory'], traj)

            # Or may there is nothing to optimise? still save a Trajectory data
            elif no_optimise is True:
                traj = TrajectoryData()
                input_structure = self.node.inputs.structure
                traj.set_trajectory(stepids=np.asarray([1]),
                                    cells=np.asarray([input_structure.cell]),
                                    symbols=np.asarray([
                                        site.kind_name
                                        for site in input_structure.sites
                                    ]),
                                    positions=np.asarray([[
                                        site.position
                                        for site in input_structure.sites
                                    ]]))
                # Save the rest
                for name, value in six.iteritems(trajectory_data):
                    # Skip saving empty arrays
                    if len(value) == 0:
                        continue

                    array = np.asarray(value)
                    # For forces/velocities we also need to resort the array
                    if ("force" in name) or ("velocities" in name):
                        array = array[:, idesort]
                    traj.set_array(name, np.asarray(value))
                self.out(out_ln['trajectory'], traj)
            # Otherwise, save data into a ArrayData node
            else:
                out_array = ArrayData()
                for name, value in six.iteritems(trajectory_data):
                    # Skip saving empty arrays
                    if len(value) == 0:
                        continue
                    array = np.asarray(value)
                    if ("force" in name) or ("velocities" in name):
                        array = array[:, idesort]
                    out_array.set_array(name, np.asarray(value))
                self.out(out_ln['array'], out_array)

        ######## ---- PROCESSING OUTPUT DATA --- ########
        output_params = Dict(dict=out_dict)
        self.out(out_ln['results'], output_params)

        # Return the exit code
        return self.exit_codes.__getattr__(exit_code)


def bands_to_bandsdata(bands_info, kpoints, bands):
    """
    Convert the result of parser_dot_bands into a BandsData object

    :param bands_info: A dictionary of the informations of the bands file.
      contains field such as eferemi, units, cell
    :param kpoints: An array of the kpoints of the bands, rows are
      (kindex, kx, ky, kz, weight)
    :param bands: The actual bands array
    :return: A BandsData object
    :rtype: ``aiida.orm.bands.data.array.bands.BandsData``
    """

    bands_node = BandsData()

    # Extract the index of the kpoints
    kpn_array = np.asarray(kpoints)
    k_index = kpn_array[:, 0]

    # We need to restore the order of the kpoints
    k_sort = np.argsort(k_index)
    # Sort the kpn_array
    kpn_array = kpn_array[k_sort]

    _weights = kpn_array[:, -1]
    kpts = kpn_array[:, 1:-1]
    bands_node.set_kpoints(kpts, weights=_weights)

    # Sort the bands to match the order of the kpoints
    bands_array = np.asarray(bands)[k_sort]
    # We need to swap the axes from kpt,spin,engs to spin,kpt,engs
    bands_array = bands_array.swapaxes(0, 1)

    # Squeeze the first dimension e.g when there is a single spin
    if bands_array.shape[0] == 1:
        bands_array = bands_array[0]
    bands_array = bands_array * units['Eh']
    bands_info['efermi'] *= units['Eh']
    bands_info['units'] = "eV"

    bands_node.set_bands(bands_array)
    bands_node.set_cell(bands_info['cell'], pbc=(True, True, True))

    # Store information from *.bands in the attributes
    # This is needs as we need to know the number of electrons
    # and the fermi energy
    for key, value in bands_info.items():
        bands_node.set_attribute(key, value)
    return bands_node



"""
Module for additional Data classes
"""

from aiida.orm.nodes.data.upf import UpfData
from aiida.common import NotExistent, MultipleObjectsError

from .otfg import OTFGData
from .usp import UspData


def get_pseudos_from_structure(structure, family_name):
    """
    Given a family name (of UpfData or OTFGData) and a AiiDA
    structure, return a dictionary associating each kind name with its
    pseduopotential object.

    :raise MultipleObjectsError: if more than one UPF for the same element is
       found in the group.
    :raise NotExistent: if no UPF for an element in the group is
       found in the group.

    :returns: A dictionary maps kind to the psueodpotential node
    """

    family_pseudos = {}

    try:
        family_upf = UpfData.get_upf_group(family_name)
    except NotExistent:
        family_upf = []
    try:
        family_otfg = OTFGData.get_otfg_group(family_name)
    except NotExistent:
        family_otfg = []

    valid_count = 0
    for fam in [family_upf, family_otfg]:
        if fam:
            valid_count += 1
            family = fam

    if valid_count == 0:
        raise NotExistent(
            "Cannot find matching group among UspData, UpfData and OTFGData")

    # This is necessary?
    if valid_count > 1:
        raise MultipleObjectsError("Name duplication detected")

    # Checking uniquess for each element
    for node in family.nodes:
        if isinstance(node, (UpfData, UspData, OTFGData)):
            if node.element in family_pseudos:
                raise MultipleObjectsError(
                    "More than one pseudo for element {} found in "
                    "family {}".format(node.element, family_name))
            family_pseudos[node.element] = node

    pseudo_list = {}
    for kind in structure.kinds:
        symbol = kind.symbol
        try:
            pseudo_list[kind.name] = family_pseudos[symbol]
        except KeyError:

            # May be we have a "LIBRARY" wild card here
            try:
                pseudo_list[kind.name] = family_pseudos["LIBRARY"]

            except KeyError:
                raise NotExistent(
                    "No pseudo for element {} found in family {}".format(
                        symbol, family_pseudos))

    return pseudo_list



"""Test module for the matchers"""
from __future__ import absolute_import
import pytest
from .raw_parser import Matcher, UnitMatcher, LineParser


def test_matcher():
    """
    Test the Marcher class
    """
    m = Matcher(r"^ +Energy is (\d+)", "energy")
    assert m.match_pattern(" Energy is 1234")[0] == ("energy", "1234")
    m = Matcher(r"^ +Energy is (\d+)", "energy", int)
    assert m.match_pattern(" Energy is 1234")[0] == ("energy", 1234)


def test_unit_matcher():
    """
    Test the UnitMatcher
    """
    m = UnitMatcher(r"^ +Energy is (\d+) +(\w+)", "energy")
    assert m.match_pattern(" Energy is 1234 eV")[0] == ("energy", float(1234),
                                                        "eV")
    m = UnitMatcher(r"^ +Energy is (\d+) +(\w+)", "energy", int)
    assert isinstance(m.match_pattern(" Energy is 1234 eV")[0][1], int)


def test_line_parser():
    """i
    Test the line parser object
    """
    m1 = UnitMatcher(r"^ +Free Energy is (\d+) +(\w+)", "free_energy")
    m2 = UnitMatcher(r"^ +Total Energy is (\d+) +(\w+)", "total_energy")
    line_parser = LineParser([m1, m2])
    res = line_parser.parse(" X Energy is 1234 eV ")
    assert res is None
    assert m2.match_pattern(" Total Energy is 1234 eV ")[0] is not None
    res = line_parser.parse(" Total Energy is 1234 eV ")
    assert res == ("total_energy", 1234, "eV")
    res = line_parser.parse(" Free Energy is 1234 Ha ")
    assert res == ("free_energy", 1234, "Ha")


def test_iter_line_parser():
    """
    Test the define line parser 
    """
    from .raw_parser import get_iter_parser

    parser = get_iter_parser()
    assert parser.parse(
        "Integrated |Spin Density|   =    0.682437E-03 hbar/2") is not None
    assert parser.parse(
        "Integrated Spin Density     =    0.202826E-03 hbar/2") is not None
    assert parser.parse(" LBFGS: finished iteration     3 with enthalpy= -2.42401562E+005 eV")[1] == \
        -2.42401562e5
    assert parser.parse(
        "NB est. 0K energy (E-0.5TS)      =  -63379.66505085   eV") is not None
    assert parser.parse(
        "Final energy, E             =  -63379.72119700     eV") is not None
    assert parser.parse(
        "Final free energy (E-TS)    =  -63374.56395158     eV") is not None



"""
Module for storing usp files into the database
"""

from __future__ import absolute_import
import os
import warnings
from aiida.plugins import DataFactory
from aiida.orm import GroupTypeString
from aiida.common.utils import classproperty
from aiida.common.files import md5_file
from .utils import get_usp_element
import six

OLD_USPGROUP_TYPE = "data.castep.usp.family"
USPGROUP_TYPE = "castep.otfg"

SinglefileData = DataFactory("singlefile")

# Extract element from filename


def upload_usp_family(folder,
                      group_label,
                      group_description,
                      stop_if_existing=True):
    """
    Upload a set of usp/recpot files in a give group

    :param folder: a path containing all UPF files to be added.
        Only files ending in .usp/.recpot are considered.
    :param group_label: the name of the group to create. If it exists and is
        non-empty, a UniquenessError is raised.
    :param group_description: a string to be set as the group description.
        Overwrites previous descriptions, if the group was existing.
    :param stop_if_existing: if True, check for the md5 of the files and,
        if the file already exists in the DB, raises a MultipleObjectsError.
        If False, simply adds the existing UPFData node to the group.
    """
    import os

    import aiida.common
    #from aiida.common import aiidalogger
    from aiida.common import UniquenessError, NotExistent
    from aiida.orm.querybuilder import QueryBuilder
    from .otfg import OTFGGroup

    files = [
        os.path.realpath(os.path.join(folder, i)) for i in os.listdir(folder)
        if os.path.isfile(os.path.join(folder, i)) and (
            i.lower().endswith('.usp') or i.lower().endswith('recpot')
            or i.lower().endswith('.uspcc'))
    ]

    nfiles = len(files)

    try:
        group = OTFGGroup.get(label=group_label)
        group_created = False
    except NotExistent:
        group = OTFGGroup(label=group_label, )
        group_created = True

    # Update the descript even if the group already existed
    group.description = group_description

    pseudo_and_created = []  # A list of records (UspData, created)

    for f in files:

        md5sum = md5_file(f)
        qb = QueryBuilder()
        qb.append(UspData, filters={'attributes.md5': {'==': md5sum}})
        existing_usp = qb.first()

        # Add the file if it is in the database
        if existing_usp is None:
            pseudo, created = UspData.get_or_create(f,
                                                    use_first=True,
                                                    store_usp=False)
            pseudo_and_created.append((pseudo, created))

        # The same file is there already
        else:
            if stop_if_existing:
                raise ValueError("A usp/recpot with identical MD5 to"
                                 " {} cannot be added with stop_if_existing"
                                 "".format(f))
            existing_usp = existing_usp[0]
            pseudo_and_created.append((existing_usp, False))

    # Check for unique per element
    elements = [(i[0].element, i[0].md5sum) for i in pseudo_and_created]

    # Check if we will duplicate after insertion

    if not group_created:
        for aiida_n in group.nodes:
            if not isinstance(aiida_n, UspData):
                continue
            elements.append((aiida_n.element, aiida_n.md5sum))

    # Discard duplicated pairs
    elements = set(elements)
    elements_names = [e[0] for e in elements]

    # Check the uniqueness of the complete group
    if not len(elements_names) == len(set(elements_names)):
        duplicates = set(
            [x for x in elements_names if elements_names.count(x) > 1])
        dup_string = ", ".join(duplicates)
        raise UniquenessError(
            "More than one usp/recpot found for the elements: " + dup_string +
            ".")

    if group_created:
        group.store()

    # Save the usp in the database if necessary and add them to the group

    for pseudo, created in pseudo_and_created:
        if created:
            pseudo.store()
            #aiidalogger.debug("New node {} created for file {}".format(
            #    pseudo.uuid, pseudo.filename))
        else:
            #aiidalogger.debug("Reusing node {} for file {}".format(
            #    pseudo.uuid, pseudo.filename))
            pass

    nodes_new = [
        pseduo for pseduo, created in pseudo_and_created if created is True
    ]
    nodes_add = [pseduo for pseduo, created in pseudo_and_created]
    group.add_nodes(nodes_add)

    return nfiles, len(nodes_new)


class UspData(SinglefileData):
    """
    Class for a single usp file
    These usp files are stored as individual file nodes in the database
    """
    def __init__(self, **kwargs):
        """
        Initialize a UspData node
        :param file str: A full path to the file of the potential
        :param element: The elemnt that this pseudo potential should be used for
        """

        element = kwargs.pop("element", None)
        self._abs_path = kwargs["file"]
        super(UspData, self).__init__(**kwargs)

        # Overides the element inferred
        if element is not None:
            self.set_element(element)

    @classmethod
    def get_or_create(cls,
                      filename,
                      element=None,
                      use_first=False,
                      store_usp=True):
        """
        Same ase init. Check md5 in the db, it is found return a UspData.
        Otherwise will store the data into the db

        :return (usp, created)
        """

        import aiida.common.utils
        import os

        # Convert the filename to an absolute path
        filename = str(filename)
        if filename != os.path.abspath(filename):
            raise ValueError("filename must be an absolute path")
        md5 = md5_file(filename)

        # Check if we have got the file already
        pseudos = cls.from_md5(md5)
        if len(pseudos) == 0:
            # No existing pseudopotential file is in the database
            instance = cls(file=filename)
            # If we there is an element given then I set it
            if element is not None:
                instance.set_element(element)
            # Store the usp if requested
            if store_usp is True:
                instance.store()
            return (instance, True)
        else:
            if len(pseudos) > 1:
                if use_first:
                    return (pseudos[0], False)
                else:
                    pks = ", ".join([str(i.pk) for i in pseudos])
                    raise ValueError("More than one copy of a pseudopotential"
                                     " found. pks={}".format(pks))
            else:
                return (pseudos[0], False)

    @classmethod
    def from_md5(cls, md5):
        """
        Return a list of all usp pseudopotentials that match a given MD5 hash.

        Note that the hash has to be stored in a md5 attribute, otherwise
        the pseudo will not be found.
        We use a special md5 attribute to avoid searching through
        irrelevant data types.
        """
        from aiida.orm.querybuilder import QueryBuilder
        qb = QueryBuilder()
        qb.append(cls, filters={'attributes.md5': {'==': md5}})
        return [_ for [_] in qb.all()]

    @classproperty
    def uspfamily_type_string(cls):
        """
        Type string of the underlying group deprecated as new 
        Group should be access by sub-classing
        """
        return USPGROUP_TYPE

    def store(self, *args, **kwargs):
        """
        Store the node. Automatically set md5 and element
        """
        # Cannot revalidate the stored nodes
        if not self.is_stored:
            self._validate()

        return super(UspData, self).store(*args, **kwargs)

    def set_file(self, filename):
        """
        Extract element and compute the md5hash
        """

        filename = str(filename)

        try:
            element = get_usp_element(filename)
        except KeyError:
            element = None
        else:
            # Only set the element if it is not there
            if self.element is None:
                if element is not None:
                    self.set_element(element)
                else:
                    warnings.warn(
                        "Cannot extract element form the usp/recpot file {}."
                        "Please set it manually.".format(filename))
            else:
                # The element is already set, no need to process further
                pass

        md5sum = md5_file(filename)
        self.set_attribute('md5', md5sum)
        super(UspData, self).set_file(filename)

    def set_element(self, element):
        """
        Set the element
        """
        self.set_attribute('element', element)

    @property
    def element(self):
        return self.get_attribute('element', None)

    @property
    def md5sum(self):
        """MD5 sum of the usp/recpot file"""
        return self.get_attribute('md5', None)

    @property
    def string(self):
        """Alias of the md5sum"""
        return self.md5sum

    @classmethod
    def get_usp_group(cls, group_label):
        """
        Return the UspFamily group with the given name.
        """
        from .otfg import OTFGGroup

        return OTFGGroup.objects.get(label=group_label)

    @classmethod
    def get_usp_groups(cls, filter_elements=None, user=None):
        """
        Return all names of groups of type UpfFamily, possibly with some filters.

        :param filter_elements: A string or a list of strings.
               If present, returns only the groups that contains one Upf for
               every element present in the list. Default=None, meaning that
               all families are returned.
        :param user: if None (default), return the groups for all users.
               If defined, it should be either a DbUser instance, or a string
               for the username (that is, the user email).
        """
        from .otfg import OTFGGroup
        from aiida.orm import QueryBuilder
        from aiida.orm import User

        query = QueryBuilder()

        query.append(OTFGGroup, tag='group', project=['*'])

        if user:
            query.append(User,
                         filters={'email': {
                             '==': user
                         }},
                         with_group='group')

        if isinstance(filter_elements, str):
            filter_elements = [filter_elements]

        if filter_elements is not None:
            actual_filter_elements = [_ for _ in filter_elements]
            query.append(
                cls,
                filters={'attributes.element': {
                    'in': filter_elements
                }},
                with_group='group')

        query.order_by({OTFGGroup: {'id': 'asc'}})
        return [_[0] for _ in query.all()]

    def _validate(self):
        from aiida.common import ValidationError

        super(UspData, self)._validate()

        # Check again, in case things changes
        usp_abspath = str(self._abs_path)

        if not usp_abspath:
            raise ValidationError("No valid usp file was passed")

        parsed_element = get_usp_element(usp_abspath)
        md5 = md5_file(usp_abspath)
        attr_element = self.element

        if attr_element is None:
            raise ValidationError("No element is set")

        attr_md5 = self.get_attribute('md5', None)
        if self.md5sum is None:
            raise ValidationError("attribute 'md5' not set.")

        if md5 != attr_md5:
            raise ValidationError(
                "Mismatch between store md5 and actual md5 value")

        # Warn if the parsed elemnt (if any) is not matching the attribute
        if attr_element != parsed_element and parsed_element is not None:
            raise ValidationError("Attribute 'element' says '{}' but '{}' was "
                                  "parsed from file name instead.".format(
                                      attr_element, parsed_element))
